"""Defines functions used to draw finite-sized boxes.

These tools are typically used to represent the extent of a voxel or
a voxel neighborhood in an image. In the context of the Point-Proposal Network,
this helps represent the region proposed by the network at layers
deeper than the original resolution of the image.

The :func:`box_trace` function is also used to represent the extent of the
active volume of the modules that make up a detector.
"""

import time

import numpy as np
import plotly.graph_objs as go

__all__ = ["scatter_boxes"]


def box_trace(
    lower,
    upper,
    draw_faces=False,
    line=None,
    linewidth=None,
    color=None,
    cmin=None,
    cmax=None,
    colorscale=None,
    intensity=None,
    hovertext=None,
    showscale=False,
    **kwargs,
):
    """Function which produces a plotly trace of a box given its lower bounds
    and upper bounds in x, y and z.

    Parameters
    ----------
    lower : np.ndarray
        (3) Vector of lower boundaries in x, z and z
    upper : np.ndarray
        (3) Vector of upper boundaries in x, z and z
    draw_faces : bool, default False
        Weather or not to draw the box faces, or only the edges
    line : dict, optional
        Dictionary which specifies box line properties
    linewidth : int, optional
        Width of the box edge lines
    color : Union[str, np.ndarray], optional
        Color of box
    cmin : float, optional
        Minimum value of the color range
    cmax : float, optional
        Maximum value of the color range
    colorscale : Union[str, dict]
        Colorscale
    intensity : Union[int, float], optional
        Color intensity of the box along the colorscale axis
    hovertext : Union[int, str, np.ndarray], optional
        Text associated with the box
    showscale : bool, default False
        If True, show the colorscale of the :class:`plotly.graph_objs.Mesh3d`
    **kwargs : dict, optional
        List of additional arguments to pass to
        :class:`plotly.graph_objs.Scatter3D` or
        :class:`plotly.graph_objs.Mesh3D`, depending on what the `draw_faces`
        parameter is set to.

    Returns
    -------
    Union[plotly.graph_objs.Scatter3D, plotly.graph_objs.Mesh3D]
        Box trace
    """
    # Check the parameters
    assert (
        len(lower) == len(upper) == 3
    ), "Must specify 3 values for both lower and upper boundaries."
    assert np.all(
        np.asarray(upper) > np.asarray(lower)
    ), "Each upper boundary should be greater than its lower counterpart."

    # List of box vertices in the edges that join them in the box mesh
    box_vertices = np.array(
        [[0, 0, 0, 0, 1, 1, 1, 1], [0, 0, 1, 1, 0, 0, 1, 1], [0, 1, 0, 1, 0, 1, 0, 1]]
    ).T
    box_edge_index = np.array(
        [[0, 0, 0, 1, 1, 2, 2, 3, 4, 4, 5, 6], [1, 2, 4, 3, 5, 3, 6, 7, 5, 6, 7, 7]]
    )
    box_tri_index = np.array(
        [
            [0, 6, 3, 2, 7, 6, 1, 1, 5, 5, 6, 7],
            [6, 0, 0, 0, 4, 4, 7, 7, 4, 0, 2, 3],
            [2, 4, 1, 3, 5, 7, 5, 3, 0, 1, 7, 2],
        ]
    )

    # List of scaled vertices
    vertices = lower + box_vertices * (upper - lower)

    # Update hovertemplate style
    hovertemplate = "x: %{x}<br>y: %{y}<br>z: %{z}"
    if hovertext is not None:
        if isinstance(hovertext, (list, tuple, np.ndarray)):
            hovertemplate += "<br>%{text}"
        else:
            hovertemplate += f"<br>{hovertext}"
            hovertext = None

    # Create the box trace
    if not draw_faces:
        # Build a list of box edges to draw (padded with None values to break
        # them from each other)
        edges = np.full((3 * box_edge_index.shape[1], 3), None)
        edges[np.arange(0, edges.shape[0], 3)] = vertices[box_edge_index[0]]
        edges[np.arange(1, edges.shape[0], 3)] = vertices[box_edge_index[1]]

        # Build a line property, if needed
        if (
            color is not None
            or linewidth is not None
            or cmin is not None
            or cmax is not None
            or colorscale is not None
        ):
            assert line is None, (
                "Must not specify `line` when providing `color`, "
                "`linewidth`, `cmin` or `cmax` independently."
            )
            if color is not None and not isinstance(color, str):
                color = np.full(len(edges), color)

            line = {
                "color": color,
                "width": linewidth,
                "cmin": cmin,
                "cmax": cmax,
                "colorscale": colorscale,
            }

        # Return trace
        trace = go.Scatter3d(
            x=edges[:, 0],
            y=edges[:, 1],
            z=edges[:, 2],
            mode="lines",
            line=line,
            hovertext=hovertext,
            hovertemplate=hovertemplate,
            **kwargs,
        )

    else:
        # If the color is a number, must be specified as an intensity
        if color is not None and not isinstance(color, str):
            assert intensity is None, "Must not provide both `color` and `intensity`."
            intensity = np.full(len(vertices), color)
            color = None

        trace = go.Mesh3d(
            x=vertices[:, 0],
            y=vertices[:, 1],
            z=vertices[:, 2],
            i=box_tri_index[0],
            j=box_tri_index[1],
            k=box_tri_index[2],
            color=color,
            intensity=intensity,
            showscale=showscale,
            cmin=cmin,
            cmax=cmax,
            colorscale=colorscale,
            hovertext=hovertext,
            hovertemplate=hovertemplate,
            **kwargs,
        )

    # Return trace
    return trace


def box_traces(
    lowers,
    uppers,
    draw_faces=False,
    color=None,
    linewidth=None,
    hovertext=None,
    cmin=None,
    cmax=None,
    shared_legend=True,
    legendgroup=None,
    showlegend=True,
    name=None,
    **kwargs,
):
    """Function which produces a list of plotly traces of boxes given a list of
    lower bounds and upper bounds in x, y and z.

    Parameters
    ----------
    lowers : np.ndarray
        (N, 3) List of vector of lower boundaries in x, z and z
    uppers : np.ndarray
        (N, 3) List of vector of upper boundaries in x, z and z
    draw_faces : bool, default False
        Weather or not to draw the box faces, or only the edges
    color : Union[str, np.ndarray], optional
        Color of boxes or list of color of boxes
    linewidth : int, default 2
        Width of the box edge lines
    hovertext : Union[int, str, np.ndarray], optional
        Text associated with every box or each box
    cmin : float, optional
        Minimum value along the color scale
    cmax : float, optional
        Maximum value along the color scale
    shared_legend : bool, default True
        If True, the plotly legend of all boxes is shared as one
    legendgroup : str, optional
        Legend group to be shared between all boxes
    showlegend : bool, default `True`
        Whether to show legends on not
    name : str, optional
        Name of the trace(s)
    **kwargs : dict, optional
        List of additional arguments to pass to
        :class:`plotly.graph_objs.Scatter3D` or
        :class:`plotly.graph_objs.Mesh3D`, depending on what the `draw_faces`
        parameter is set to.

    Returns
    -------
    Union[List[plotly.graph_objs.Scatter3D], List[plotly.graph_objs.Mesh3D]]
        Box traces
    """
    # Check the parameters
    assert len(lowers) == len(
        uppers
    ), "Provide as many upper boundary vector as their lower counterpart."
    assert (
        color is None or np.isscalar(color) or len(color) == len(lowers)
    ), "Specify one color for all boxes, or one color per box."
    assert (
        hovertext is None or np.isscalar(hovertext) or len(hovertext) == len(lowers)
    ), "Specify one hovertext for all boxes, or one hovertext per box."

    # If one color is provided per box, give an associated hovertext
    if hovertext is None and isinstance(color, (list, tuple, np.ndarray)):
        hovertext = [f"Value: {v:0.3f}" for v in color]

    # If cmin/cmax are not provided, must build them so that all boxes
    # share the same colorscale range (not guaranteed otherwise)
    if isinstance(color, (list, tuple, np.ndarray)) and len(color) > 0:
        if cmin is None:
            cmin = np.min(color)
        if cmax is None:
            cmax = np.max(color)

    # If the legend is to be shared, make sure there is a common legend group
    if shared_legend and legendgroup is None:
        legendgroup = "group_" + str(time.time())

    # Loop over the list of box boundaries
    traces = []
    for i, (lower, upper) in enumerate(zip(lowers, uppers)):
        # Fetch the right color/hovertext combination
        col, hov = color, hovertext
        if isinstance(color, (list, tuple, np.ndarray)):
            col = color[i]
        if isinstance(hovertext, (list, tuple, np.ndarray)):
            hov = hovertext[i]

        # If the legend is shared, only draw the legend of the first trace
        if shared_legend:
            showlegend = showlegend and i == 0
            name_i = name
        else:
            name_i = f"{name} {i}"

        # Append list of traces
        traces.append(
            box_trace(
                lower,
                upper,
                draw_faces,
                linewidth=linewidth,
                color=col,
                hovertext=hov,
                cmin=cmin,
                cmax=cmax,
                legendgroup=legendgroup,
                showlegend=showlegend,
                name=name_i,
                **kwargs,
            )
        )

    return traces


def scatter_boxes(
    coords,
    dimension,
    draw_faces=True,
    color="orange",
    hovertext=None,
    linewidth=2,
    shared_legend=True,
    **kwargs,
):
    """Function which produces a list of plotly traces of boxes given a list of
    coordinates and a box dimension.

    This function assumes that the coordinates represent the lower bounds of
    the voxels they point at. This follows the `MinkowskiEngine` convention,
    which is the package used for space convolutions. This can be used to
    represent the PPN regions of interest in a space compressed by a factor
    (b_x, b_y, b_z) from the original image resolution.

    Parameters
    ----------
    coords : np.ndarray
        (N, 3) Coordinates of in multiples of box lengths in each dimension
    dimension : Union[float, np.ndarray]
        Dimensions of the boxes. Specify it as either a single number (for
        cubes) or an array of values in each dimension, i.e. (b_x, b_y, b_z)
    draw_faces : bool, default True
        Weather or not to draw the box faces, or only the edges
    color : Union[str, np.ndarray], default 'orange'
        Color of boxes or list of color of boxes
    hovertext : Union[int, str, np.ndarray], optional
        Text associated with every box or each box
    linewidth : int, default 2
        Width of the box edge lines
    shared_legend : bool, default True
        If True, the plotly legend of all boxes is shared as one
    **kwargs : dict, optional
        List of additional arguments to pass to
        :class:`plotly.graph_objs.Scatter3D` or
        :class:`plotly.graph_objs.Mesh3D`, depending on what the `draw_faces`
        parameter is set to.

    Returns
    -------
    Union[List[plotly.graph_objs.Scatter3D], List[plotly.graph_objs.Mesh3D]]
        Box traces
    """
    # Check the input
    if isinstance(dimension, (list, tuple, np.ndarray)):
        assert len(dimension) == 3, "Must specify three dimensions for the box size."
        dimension = np.asarray(dimension)

    # Compute the lower and upper boundaries, return box traces
    lowers = coords
    uppers = coords + dimension

    return box_traces(
        lowers, uppers, draw_faces, color, linewidth, hovertext, shared_legend, **kwargs
    )



"""Module of functions that approximate the energy loss of particles through
matter. It includes function to compute the CSDA KE of particles given their
range and vice-versa.
"""

import os
import pathlib

import numba as nb
import numpy as np
import pandas as pd
from scipy.constants import fine_structure
from scipy.integrate import quad
from scipy.interpolate import CubicSpline
from scipy.optimize import brentq
from scipy.special import digamma

from .globals import (
    ELEC_MASS,
    KAON_PID,
    LAR_A,
    LAR_DENSITY,
    LAR_MEE,
    LAR_Z,
    MUON_MASS,
    MUON_PID,
    PION_PID,
    PROT_PID,
    LAR_a,
    LAR_Cbar,
    LAR_delta0,
    LAR_k,
    LAR_x0,
    LAR_x1,
)


def csda_table_spline(particle_type, value="T", table_dir="csda_tables"):
    """Interpolates a CSDA table to form a spline which maps a range to a
    kinematic energy estimate.

    Parameters
    ----------
    particle_type : int
        Particle type ID to construct splines. Maps are
        avaible for muons, pions, kaons and protons.
    value : str, default 'T'
        Value to provide for each range value (one of 'T' or 'dE/dx')
    table_dir : str, default 'csda_tables'
        Relative path to the CSDA range tables

    Returns
    -------
    callable
        Function mapping range (cm) to Kinetic E (MeV) or dE/dx (MeV/cm)
    """
    # Assert that the request value is recognized
    assert value in [
        "T",
        "dE/dx",
    ], "Can only provide the kinetic energy or dE/dx for a given range."
    factor = 1.0 if value == "T" else LAR_DENSITY

    # Check that the table for the requested PID exists
    path = pathlib.Path(__file__).parent
    suffix = "E_liquid_argon"
    name_mapping = {MUON_PID: "mu", PION_PID: "pi", KAON_PID: "ka", PROT_PID: "p"}
    if particle_type not in name_mapping.keys():
        raise ValueError(
            f"CSDA table for particle type {particle_type} is not available."
        )

    # Fetch the table and fit a spline
    pid = name_mapping[particle_type]
    file_name = os.path.join(path, table_dir, f"{pid}{suffix}")
    if os.path.isfile(f"{file_name}.table"):
        path = f"{file_name}.table"
    else:
        path = f"{file_name}_bethe.table"

    tab = pd.read_csv(path, delimiter=" ", index_col=False)
    f = CubicSpline(tab["CSDARange"] / LAR_DENSITY, tab[value] * factor)

    return f


def csda_ke_lar(R, M, z=1, T_max=1e6, epsrel=1e-3, epsabs=1e-3):
    """Numerically optimizes the kinetic energy necessary to observe the
    range of a particle that has been measured, under the CSDA.

    Parameters
    ----------
    R : float
        Range that the particle travelled through liquid argon in cm
    M : float
        Particle mass in MeV/c^2
    z : int, default 1
        Impinging partile charge in multiples of electron charge
    T_max : float, default 1e6
        Maximum allowed kinetic energy
    epsrel : float, default 1e-3
        Relative error tolerance
    epsabs : float, default 1e-3
        Asbolute error tolerance

    Returns
    -------
    float
        CSDA kinetic energy in MeV
    """
    func = lambda x: csda_range_lar(x, M, z, epsrel, epsabs) - R
    return brentq(func, 0.0, T_max, rtol=epsrel, xtol=epsabs)


def csda_range_lar(T0, M, z=1, epsrel=1e-3, epsabs=1e-3):
    """Numerically integrates the inverse Bethe-Bloch formula to find the
    CSDA range of a particle for a given initial kinetic energy.

    Parameters
    ----------
    T0 : float
        Initial kinetic energy in MeV
    M : float
        Particle mass in MeV/c^2
    z : int, default 1
        Impinging partile charge in multiples of electron charge
    epsrel : float, default 1e-3
        Relative error tolerance
    epsabs : float, default 1e-3
        Asbolute error tolerance

    Returns
    -------
    float
        CSDA range in cm
    """
    if T0 <= 0.0:
        return 0.0

    return -quad(
        inv_bethe_bloch_lar, 0.0, T0, args=(M, z), epsrel=epsrel, epsabs=epsabs
    )[0]


@nb.njit(cache=True)
def step_energy_loss_lar(T0, M, dx, z=1, num_steps=None):
    """Steps the initial energy of a particle down by pushing it through
    steps of dx of liquid argon. If `num_steps` is not specified, it will
    iterate until the particle kinetic energy reaches 0.

    Parameters
    ----------
    T0 : float
        Initial kinetic energy in MeV
    M : float
        Particle mass in MeV/c^2
    dx : float
        Step size in cm
    z : int, default 1
        Impinging partile charge in multiples of electron charge
    num_steps : int, optional
        If specified, only step a maximum of `num_steps` times

    Returns
    -------
    np.array
        Array of kinetic energies at each step
    """
    # Initialize the list
    assert T0 > 0.0, "Must start with positive kinetic energy"
    ke_list = [T0]

    # Step down
    step = 0
    Ti = T0
    while Ti > 0 and (step is None or step < num_steps):
        step += 1
        Ti += dx * bethe_bloch_lar(Ti, M, z)
        if Ti > 0.0:
            ke_list.append(Ti)
        else:
            ke_list.append(0.0)
            break

    # Return
    return np.array(ke_list)


@nb.njit(cache=True)
def inv_bethe_bloch_lar(T, M, z=1):
    """Inverse Bethe-Bloch energy loss function for liquid argon.

    Parameters
    ----------
    T : float
       Kinetic energy in MeV
    M : float
       Impinging particle mass in MeV/c^2
    z : int, default 1
       Impinging partile charge in multiples of electron charge

    Returns
    -------
    float
       Value of the inverse energy loss rate in liquid argon in MeV/cm
    """
    return 1.0 / bethe_bloch_lar(T, M, z)


@nb.njit(cache=True)
def bethe_bloch_lar(T, M, z=1):
    """Bethe-Bloch energy loss function for liquid argon.

    Reference:

    https://pdg.lbl.gov/2019/reviews/rpp2018-rev-passage-particles-matter.pdf

    Corrections taken from:

    https://pdg.lbl.gov/2023/AtomicNuclearProperties/adndt.pdf

    Parameters
    ----------
    T : float
       Kinetic energy in MeV
    M : float
       Impinging particle mass in MeV/c^2
    z : int, default 1
       Impinging partile charge in multiples of electron charge

    Returns
    -------
    float
       Value of the energy loss rate in liquid argon in MeV/cm
    """
    # Constant
    K = 0.307075  # Bethe-Bloch constant [MeV*cm^2/mol]

    # Kinematics
    gamma = 1.0 + T / M
    beta = np.sqrt(1.0 - 1.0 / gamma**2)
    bg = beta * gamma

    # Prefactor
    F = -K * z**2 * (LAR_Z / LAR_A) * LAR_DENSITY / beta**2

    # Compute the max energy transfer
    W = W_max(beta, gamma, M)

    # Compute the density effects
    delta = delta_lar(bg)

    # Compute the low energy corrections
    le_corr = 0.0  # le_corr_lar(beta, z)

    # Compute the Bremsstrahlung correction
    del_dedx = (
        -K
        * fine_structure
        * (LAR_Z / LAR_A)
        / (4 * np.pi)
        * (np.log(2 * gamma) - 1.0 / 3 * (np.log(2 * W / ELEC_MASS)))
        * np.log(2 * W / ELEC_MASS) ** 2
    )

    # Compute the muon-specific spin correction
    spin_corr_muon = (1.0 / 8) * (W / gamma / M) ** 2 * (M == MUON_MASS)

    return (
        F
        * (
            0.5 * np.log((2 * ELEC_MASS * bg**2 * W) / LAR_MEE**2)
            - beta**2
            - 0.5 * delta
            + le_corr
            + spin_corr_muon
        )
        + del_dedx
    )


@nb.njit(cache=True)
def bethe_bloch_mpv_lar(T, M, x, z=1):
    """Most-probable value of energy loss through a thin layer of liquid argon.

    https://pdg.lbl.gov/2019/reviews/rpp2018-rev-passage-particles-matter.pdf

    Parameters
    ----------
    T : float
       Kinetic energy in MeV
    M : float
       Impinging particle mass in MeV/c^2
    x : float
       Material thickness in cm
    z : int, default 1
       Impinging partile charge in multiples of electron charge

    Returns
    -------
    float
       Value of the energy loss in liquid argon in MeV
    """
    # Constants
    K = 0.307075  # Bethe-Bloch constant [MeV*cm^2/mol]
    j = 0.200

    # Kinematics
    gamma = 1.0 + T / M
    beta = np.sqrt(1.0 - 1.0 / gamma**2)
    bg = beta * gamma

    # Prefactor
    xi = (K / 2) * z**2 * (LAR_Z / LAR_A) * LAR_DENSITY * x / beta**2

    # Compute the density effects
    delta = delta_lar(bg)

    return -xi * (
        np.log(2 * ELEC_MASS * (bg**2) / LAR_MEE)
        + np.log(xi / LAR_MEE)
        + j
        - beta**2
        - delta
    )


@nb.njit(cache=True)
def W_max(beta, gamma, M):
    """Maximum energy transfer in a single colision.

    Parameters
    ----------
    beta : float
        Lorentz beta (v/c)
    gamma : float
        Lorentz gamma (1/sqrt(1-beta**2))
    M : float
        Mass of the impinging particle in MeV/c^2

    Returns
    -------
    float
        Maximum energy transferred in a single colision
    """
    bg = beta * gamma
    return (2 * ELEC_MASS * bg**2) / (
        1.0 + 2 * gamma * ELEC_MASS / M + (ELEC_MASS / M) ** 2
    )


@nb.njit(cache=True)
def delta_lar(bg):
    """
    Density correction

    Parameters
    ----------
    bg : float
        Product of Lorentz beta and gamma (beta/sqrt(1-beta**2))

    Returns
    -------
    float
        Density correction to the Bethe-Bloch function
    """
    x = np.log10(bg)
    if x < LAR_x0:
        return LAR_delta0 * 10 ** (2 * (x - LAR_x0))
    elif x >= LAR_x0 and x < LAR_x1:
        return 2 * np.log(10) * x - LAR_Cbar + LAR_a * (LAR_x1 - x) ** LAR_k
    else:
        return 2 * np.log(10) * x - LAR_Cbar


# @nb.njit(cache=True) # Find an alternative to scipy's digamma to support njit
def le_corr_lar(beta, z=1):
    """Low energy corrections to the Bethe-Bloch formula.

    Parameters
    ----------
    beta : float
        Lorentz beta (v/c)
    z : int, default 1
        Impinging partile charge in multiples of electron charge

    Returns
    -------
    float
        Low energy correction to the energy loss function
    """
    # Shell corrections (tabulated, ignored for now)
    C = 0.0

    # Barkas Correction (tabulated, ignored for now)
    L1 = 0.0

    # Bloch Correction
    y = fine_structure * z / beta
    L2 = -abs(y) - np.real(digamma(1 + y * 1j))

    # Low energy correction
    return -C / LAR_Z + z * L1 + z**2 * L2



"""Module which defines encoders that mix geometric and CNN features together.

See :mod:`spine.model.layer.gnn.encode.geometric` and
:mod:`spine.model.layer.gnn.encode.cnn` for more information.
"""

import torch

from spine.model.layer.common.act_norm import act_factory

from .cnn import ClustCNNEdgeEncoder, ClustCNNNodeEncoder
from .geometric import ClustGeoEdgeEncoder, ClustGeoNodeEncoder

__all__ = ["ClustGeoCNNMixNodeEncoder", "ClustGeoCNNMixEdgeEncoder"]


class ClustGeoCNNMixNodeEncoder(torch.nn.Module):
    """Produces cluster node features using both geometric and CNN encoders."""

    # Name of the node encoder (as specified in the configuration)
    name = "geo_cnn_mix"

    def __init__(self, geo_encoder, cnn_encoder, activation="elu"):
        """Initialize the mixed encoder.

        Initializes the two underlying encoders:
        - :class:`ClustGeoNodeEncoder`
        - :class:`ClustCNNNodeEncoder`

        Parameters
        ----------
        geo_encoder : dict
            Geometric node encoder configuration
        cnn_encoder : dict,
            CNN node encoder configuration
        """
        # Initialize the parent class
        super().__init__()

        # Initialize the two underlying encoder
        self.geo_encoder = ClustGeoNodeEncoder(**geo_encoder)
        self.cnn_encoder = ClustCNNNodeEncoder(**cnn_encoder)

        # Get the number of features coming out of the encoder
        num_geo = self.geo_encoder.feature_size
        num_cnn = self.cnn_encoder.feature_size
        self.feature_size = num_geo + num_cnn

        # Initialize the final activation/linear layer
        self.act = act_factory(activation)
        self.linear = torch.nn.Linear(self.features_size, self.feature_size)

    def forward(self, data, clusts, **kwargs):
        """Generate mixed cluster node features for one batch of data.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) Batch of sparse tensors
        clusts : IndexBatch
            (C) List of list of indexes that make up each cluster
        **kwargs : dict, optional
            Additional objects no used by this encoder

        Returns
        -------
        TensorBatch
            (C, N_c) Set of N_c features per cluster
        """
        # Get the features
        features_geo = self.geo_encoder(data, clusts, **kwargs).tensor
        features_cnn = self.cnn_encoder(data, clusts, **kwargs).tensor
        features_mix = torch.cat([features_geo, features_cnn], dim=1)

        # Push features through the final activation/linear layer
        result = self.act(features_mix)
        result = self.linear(result)

        return TensorBatch(result, clusts.list_counts)


class ClustGeoCNNMixEdgeEncoder(torch.nn.Module):
    """Produces cluster edge features using both geometric and CNN encoders."""

    # Name of the edge encoder (as specified in the configuration)
    name = "geo_cnn_mix"

    def __init__(self, geo_encoder, cnn_encoder):
        """Initialize the mixed encoder.

        Initializes the two underlying encoders:
        - :class:`ClustGeoEdgeEncoder``
        - :class:`ClustCNNEdgeEncoder`

        Parameters
        ----------
        geo_encoder : dict
            Geometric edge encoder configuration
        cnn_encoder : dict,
            CNN edge encoder configuration
        """
        # Initialize the parent class
        super().__init__()

        # Initialize the two underlying encoder
        self.geo_encoder = ClustGeoEdgeEncoder(**geo_encoder)
        self.cnn_encoder = ClustCNNEdgeEncoder(**cnn_encoder)

        # Get the number of features coming out of the encoder
        num_geo = self.geo_encoder.feature_size
        num_cnn = self.cnn_encoder.feature_size
        self.feature_size = num_geo + num_cnn

        # Initialize the final activation/linear layer
        self.act = act_factory(activation)
        self.linear = torch.nn.Linear(self.features_size, self.feature_size)

    def forward(self, data, clusts, edge_index, **kwargs):
        """Generate mixed cluster edge features for one batch of data.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) Batch of sparse tensors
        clusts : IndexBatch
            (C) List of list of indexes that make up each cluster
        edge_index : EdgeIndexBatch
            Incidence map between clusters
        **kwargs : dict, optional
            Additional objects no used by this encoder

        Returns
        -------
        TensorBatch
            (C, N_e) Set of N_e features per edge
        """
        # Get the features
        features_geo = self.geo_encoder(data, clusts, **kwargs).tensor
        features_cnn = self.cnn_encoder(data, clusts, **kwargs).tensor
        features_mix = torch.cat([features_geo, features_cnn], dim=1)

        # Push features through the final activation/linear layer
        result = self.act(features_mix)
        result = self.linear(result)

        return TensorBatch(result, edge_index.counts)



import numba as nb
import numpy as np

import spine.math as sm

from .globals import INTER_COL, PRINT_COL, TRACK_SHP, VTX_COLS


def get_vertex(
    start_points,
    end_points,
    directions,
    semantics,
    anchor_vertex=True,
    touching_threshold=2.0,
    return_mode=False,
):
    """Reconstruct the vertex of an individual interaction.

    Parameters
    ----------
    start_points : np.ndarray
        (P, 3) Particle start points
    end_points : np.ndarray
        (P, 3) Particle end points
    directions : np.ndarray
        (P, 3) Particle directions
    semantics : np.ndarray
        (P) : particle semantic type
    anchor_vertex : bool, default True
        If true, anchor the candidate vertex to particle objects,
        with the expection of interactions only composed of showers.
    touching_threshold : float, default 2.0
        Maximum distance for two particle points to be considered touching
    return_mode : bool, default False
        If `True`, return the method used to find the vertex
    """
    # If there is no particle: return default values
    if not len(start_points):
        if return_mode:
            return np.full(3, -np.inf), "no_particle"
        return np.full(3, -np.inf)

    # If there is only one particle: choose the start point
    if len(start_points) == 1:
        if return_mode:
            return start_points[0], "single_start"
        return start_points[0]

    # If this more than one particle, anchor the vertex to a particle point
    track_mask = semantics == TRACK_SHP
    track_ids = np.where(track_mask)[0]
    if anchor_vertex:
        # If there is a unique point where >=2 particles meet, pick it. Include
        # track start and end points, to not rely on direction predictions
        vertices = get_confluence_points(start_points, end_points, touching_threshold)
        if len(vertices) == 1:
            if return_mode:
                return vertices[0], "confluence_nodir"
            return vertices[0]

        # If there is more than one option, restrict track end points to the
        # predicted start points (relies on direction prediction), check again
        if len(vertices) > 1 and len(track_ids):
            vertices = get_confluence_points(
                start_points, touching_threshold=touching_threshold
            )
            if len(vertices) == 1:
                if return_mode:
                    return vertices[0], "confluence_dir"
                return vertices[0]

        # If there are no obvious confluence points, but there is a single
        # track and N showers, pick the track end-point which minimizes the
        # cosine distance between the normalized vertex-to-point vectors
        # and the shower direction estimates.
        if len(track_ids) == 1:
            candidates = np.vstack([start_points[track_mask], end_points[track_mask]])
            losses = angular_loss(
                candidates, start_points[~track_mask], directions[~track_mask]
            )
            if return_mode:
                return candidates[np.argmin(losses)], "confluence_track_showers"
            return candidates[np.argmin(losses)]

        # If all else fails on track groups, simply pick the longest
        # track, take its starting point and hope for the best...
        if len(track_ids):
            track_lengths = np.linalg.norm(
                end_points[track_mask] - start_points[track_mask], axis=-1
            )
            if return_mode:
                return (
                    start_points[track_mask][np.argmax(track_lengths)],
                    "track_length",
                )
            return start_points[track_mask][np.argmax(track_lengths)]

    # If there is only showers (or the vertex is not anchored): find the point which
    # minimizes the sum of distances from the vertex w.r.t. to all direction lines.
    # TODO: Find the point which minimizes the angular difference between the
    # vertex-to-point vector and the shower direction estimates (much better).
    if return_mode:
        return get_pseudovertex(start_points, directions), "pseudo_vertex"
    return get_pseudovertex(start_points, directions)


@nb.njit(cache=True)
def angular_loss(
    candidates: nb.float32[:, :],
    points: nb.float32[:, :],
    directions: nb.float32[:, :],
    use_cos: bool = True,
) -> nb.float32:
    """Computes the angular/cosine distance between vectors that join candidate
    points to the start points of particles and their respective direction
    estimates. Values are normalized between 0 (perfect fit) and 1
    (complete disagreement).

    Parameters
    ----------
    candidates : np.ndarray
        (C, 3) Vertex coordinates
    points : np.ndarray
        (P, 3) Particle start points
    directions : np.ndarray
        (P, 3) Particle directions
    use_cos : bool
        Whether or not to use the cosine as a metric

    Returns
    -------
    np.ndarray
        (C) Loss for each of the candidates
    """
    n_c = len(candidates)
    losses = np.empty(n_c, dtype=np.float32)
    for i, c in enumerate(candidates):
        loss = 0.0
        for p, d in zip(points, directions):
            v = p - c
            v /= np.linalg.norm(v)
            if use_cos:
                loss += (1.0 - np.sum(v * d)) / 2 / n_c
            else:
                loss += np.arccos(np.sum(v * d)) / np.pi / n_c

        losses[i] = loss

    return losses


@nb.njit(cache=True)
def get_confluence_points(
    start_points: nb.float32[:, :],
    end_points: nb.float32[:, :] = None,
    touching_threshold: nb.float32 = 2.0,
) -> nb.types.List(nb.float32[:]):
    """Find the points where multiple particles touch.

    Parameters
    ----------
    start_points : np.ndarray
        (P, 3) Particle start points
    end_points : np.ndarray, optional
        (P, 3) Particle end points
    touching_threshold : float, default 2.0
        Maximum distance for two particle points to be considered touching

    Returns
    -------
    List[np.ndarray]
        List of vertices that correspond to the confluence points
    """
    # Create a particle-to-particle distance matrix
    n_part = len(start_points)
    dist_mat = np.zeros((n_part, n_part), dtype=start_points.dtype)
    end_mat = np.zeros((n_part, n_part), dtype=np.int32)
    if end_points is None:
        for i, si in enumerate(start_points):
            for j, sj in enumerate(start_points):
                if j > i:
                    dist_mat[i, j] = np.linalg.norm(sj - si)
                if j < i:
                    dist_mat[i, j] = dist_mat[j, i]

    else:
        for i, (si, ei) in enumerate(zip(start_points, end_points)):
            pointsi = np.vstack((si, ei))
            for j, (sj, ej) in enumerate(zip(start_points, end_points)):
                if j > i:
                    pointsj = np.vstack((sj, ej))
                    submat = sm.distance.cdist(pointsi, pointsj)
                    mini, minj = np.argmin(submat) // 2, np.argmin(submat) % 2
                    dist_mat[i, j] = submat[mini, minj]
                    end_mat[i, j], end_mat[j, i] = mini, minj
                if j < i:
                    dist_mat[i, j] = dist_mat[j, i]

    # Convert distance matrix to an adjacency matrix, compute
    # the square graphic to find the number of walks
    adj_mat = (dist_mat < touching_threshold).astype(
        np.float32
    )  # @ does not like integers
    walk_mat = adj_mat @ adj_mat
    for i in range(n_part):
        walk_mat[i, i] = np.max(walk_mat[i][np.arange(n_part) != i])

    # Find cycles to build particle groups and confluence points (vertices)
    leftover = np.ones(n_part, dtype=np.bool_)
    max_walks = sm.amax(walk_mat, axis=1)
    vertices = nb.typed.List.empty_list(np.empty(0, dtype=start_points.dtype))
    while np.any(leftover):
        # Find the longest available cycle (must be at least 2 particles)
        left_ids = np.where(leftover)[0]
        max_id = left_ids[np.argmax(max_walks[leftover])]
        max_walk = max_walks[max_id]
        if max_walk < 2:
            break

        # Form the particle group that make up the cycle
        leftover[walk_mat[max_id] == max_walk] = False
        group = np.where(walk_mat[max_id] == max_walk)[0]

        # Take the barycenter of the touching particle ends as the vertex
        if end_points is None:
            vertices.append(sm.mean(start_points[group], axis=0))
        else:
            vertex = np.zeros(3, dtype=start_points.dtype)
            for i, t in enumerate(group):
                end_id = np.argmax(
                    np.bincount(end_mat[t][group][np.arange(len(group)) != i])
                )
                vertex += (
                    start_points[t] / len(group)
                    if not end_id
                    else end_points[t] / len(group)
                )
            vertices.append(vertex)

    return vertices


@nb.njit(cache=True)
def get_pseudovertex(
    start_points: nb.float32[:, :], directions: nb.float32[:, :], dim: int = 3
) -> nb.float32[:]:
    """Finds the vertex which minimizes the total distance from itself to all
    the lines defined by the start points of particles and their directions.

    Parameters
    ----------
    start_points : np.ndarray
        (P, 3) Particle start points
    directions : np.ndarray
        (P, 3) Particle directions
    dim : int
        Number of dimensions
    """
    assert len(start_points), "Cannot reconstruct pseudovertex without points."

    if len(start_points) == 1:
        return start_points[0]

    if np.all((directions[1:] == directions[0]) | (directions[1:] == -directions[0])):
        return sm.mean(start_points, axis=0)

    S = np.zeros((dim, dim), dtype=start_points.dtype)
    C = np.zeros((dim,), dtype=start_points.dtype)
    for p, d in zip(start_points, directions):
        x = np.outer(d, d) - np.eye(dim, dtype=start_points.dtype)
        S += x
        C += x @ np.ascontiguousarray(p)

    # TODO remove casting to float64 when the crash goes away
    return np.linalg.inv(S.astype(np.float64)).astype(start_points.dtype) @ C


@nb.njit(cache=True)
def get_weighted_pseudovertex(
    start_points: nb.float32[:, :],
    directions: nb.float32[:, :],
    weights: nb.float32[:],
    dim: int = 3,
) -> nb.float32[:]:
    """Finds the vertex which minimizes the total distance from itself to all
    the lines defined by the start points of particles and their directions.

    Parameters
    ----------
    start_points : np.ndarray
        (P, 3) Particle start points
    directions : np.ndarray
        (P, 3) Particle directions
    dim : int
        Number of dimensions
    """
    assert len(start_points), "Cannot reconstruct pseudovertex without points."

    if len(start_points) == 1:
        return start_points[0]

    if np.all(directions[1:] == directions[0] | directions[1:] == -directions[0]):
        return sm.sum(weights * start_points, axis=0) / np.sum(weights)

    S = np.zeros((dim, dim), dtype=start_points.dtype)
    C = np.zeros((dim,), dtype=start_points.dtype)
    for i, (p, d) in enumerate(zip(start_points, directions)):
        x = weights[i] * (np.outer(d, d) - np.eye(dim, dtype=start_points.dtype))
        S += x
        C += x @ np.ascontiguousarray(p)

    # TODO remove casting to float64 when the crash goes away
    # TODO check for singular matrix or go back to pinv when fixed
    return np.linalg.inv(S.astype(np.float64)).astype(start_points.dtype) @ C



"""Module to convert a point cloud into an cone envelope."""

import numpy as np
from plotly import graph_objs as go

from spine.math.decomposition import principal_components

__all__ = ["cone_trace"]


def cone_trace(
    points,
    fraction=0.5,
    num_samples=10,
    color=None,
    hovertext=None,
    intensity=None,
    showscale=False,
    **kwargs,
):
    """Converts a cloud of points into a 3D cone.

    This function uses the PCA and the average angle w.r.t. to the point
    of maximum curvature as a basis to construct a cone.

    Parameters
    ----------
    points : np.ndarray
        (N, 3) Array of point coordinates
    fraction : float, default 0.5
        Fraction of the points contained in the cone (angle quantile)
    num_samples : int, default 10
        Number of points sampled along h and phi in the conical coordinate
        system of the cone. A larger number increases the resolution.
    showscale : bool, default False
        If True, show the colorscale of the :class:`plotly.graph_objs.Mesh3d`
    color : Union[str, float], optional
        Color of the cone
    hovertext : Union[int, str, np.ndarray], optional
        Text associated with the cone
    intensity : Union[str, float, np.ndarray], optional
        Intensity of the cone colors
    showscale : bool, default False
        Whether to show the color scale for the mesh
    **kwargs : dict, optional
        Additional parameters to pass to the :class:`plotly.graph_objs.Mesh3d` object
    """
    # Get the centroid and the principal components
    centroid = np.mean(points, axis=0)
    pcomp = principal_components(points)
    rotmat = np.flip(pcomp, axis=0)
    paxis = pcomp[0]

    # Collapse point cloud onto principal axis, find end points
    points_pa = np.dot(points, paxis)
    end_ids = np.argmin(points_pa), np.argmax(points_pa)

    # Find the directions w.r.t. each end point
    centroid = np.mean(points, axis=0)
    dirs = np.vstack([centroid - points[end_id] for end_id in end_ids])
    dirs = dirs / np.linalg.norm(dirs, axis=1)[:, None]

    # Find a quantile angle w.r.t. to each direction
    dots = np.zeros((2, len(points)), dtype=points.dtype)
    for i, end_id in enumerate(end_ids):
        for j, point in enumerate(points):
            if j != end_id:
                diff = point - points[end_id]
                dots[i, j] = np.dot(diff / np.linalg.norm(diff), dirs[i])

    assert (
        fraction > 0.0 and fraction < 1.0
    ), "The `fraction` parameter should be a probability."
    angles = np.arccos(dots)
    means = np.mean(angles, axis=1)
    quantiles = np.quantile(angles, fraction, axis=1)

    # The point with the lowest mean angle is the start, select it
    start_id = np.argmin(means)
    start_pos = points[end_ids[start_id]]

    # Define the cone main axis length and its opening angle
    length = abs(points_pa[end_ids[1]] - points_pa[end_ids[0]])
    theta = quantiles[start_id]

    # Compute the points on a cone with half-opening angle theta
    r = np.linspace(0, 1, num=num_samples)
    phi = np.linspace(0, 2 * np.pi, num=num_samples)
    r, phi = np.meshgrid(r, phi)
    x = r * np.tan(theta) * np.cos(phi)
    y = r * np.tan(theta) * np.sin(phi)
    z = r
    unit_points = np.vstack((x.flatten(), y.flatten(), z.flatten())).T

    # Rotate and offset the cone
    cone_points = start_pos + length * np.dot(unit_points, rotmat)

    # Convert the color provided to a set of intensities
    if color is not None:
        assert intensity is None, "Provide either `color` or `intensity`, not both."
        assert np.isscalar("color"), "Should provide a single color for the cone."
        intensity = [color] * len(cone_points)

    # Update hovertemplate style
    hovertemplate = "x: %{x}<br>y: %{y}<br>z: %{z}"
    if hovertext is not None:
        if not np.isscalar(hovertext):
            hovertemplate += "<br>%{text}"
        else:
            hovertemplate += f"<br>{hovertext}"
            hovertext = None

    # Append Mesh3d object
    return go.Mesh3d(
        x=cone_points[:, 0],
        y=cone_points[:, 1],
        z=cone_points[:, 2],
        intensity=intensity,
        alphahull=0,
        showscale=showscale,
        hovertemplate=hovertemplate,
        **kwargs,
    )



"""Module with functions that operate on single voxels in the context of GNNs."""

import numba as nb
import numpy as np

import spine.math as sm
from spine.data import TensorBatch
from spine.utils.globals import COORD_COLS
from spine.utils.jit import numbafy


def get_voxel_features_batch(data, max_dist=5.0):
    """Returns an array of features for each voxel.

    The basic 16 geometric features are composed of:
    - Voxel coordinates
    - Covariance matrix of its neighborhood (3)
    - Principal axis of its neighborhood (3)
    - Voxel count in its neighborhood (1)

    The neighborhood of the voxel i defined as all voxels within some distance
    of the voxel to get features for.

    Parameters
    ----------
    data : TensorBatch
        Batch of cluster label data tensor
    max_dist : float, default 5.0
        Neighborhood radius

    Returns
    -------
    np.ndarray
        (C, N_c) Tensor of voxels features
    """
    feats = get_voxel_features(data.tensor, max_dist)

    return TensorBatch(feats, data.counts)


@numbafy(cast_args=["data"], keep_torch=True, ref_arg="data")
def get_voxel_features(data, max_dist=5.0):
    """Returns an array of features for each voxel.

    The basic 16 geometric features are composed of:
    - Voxel coordinates
    - Covariance matrix of its neighborhood (3)
    - Principal axis of its neighborhood (3)
    - Voxel count in its neighborhood (1)

    The neighborhood of the voxel i defined as all voxels within some distance
    of the voxel to get features for.

    Parameters
    ----------
    data : np.ndarray
        Cluster label data tensor
    max_dist : float, default 5.0
        Neighborhood radius

    Returns
    -------
    np.ndarray
        (C, N_c) Tensor of voxels features
    """
    return _get_voxel_features(voxels, max_dist)


@nb.njit(parallel=True, cache=True)
def _get_voxel_features(data: nb.float32[:, :], max_dist=5.0):

    # Compute intervoxel distance matrix
    voxels = data[:, COORD_COLS]
    dist_mat = sm.distance.cdist(voxels, voxels)

    # Get local geometrical features for each voxel
    feats = np.empty((len(voxels), 16), dtype=data.dtype)
    for k in nb.prange(len(voxels)):

        # Restrict the points to the neighborood of the voxel
        voxel = voxels[k]
        x = voxels[dist_mat[k] < max_dist]

        # Get orientation matrix
        A = np.cov(x.T, ddof=len(x) - 1).astype(x.dtype)

        # Center data around voxel
        x = x - voxel

        # Get eigenvectors, normalize orientation matrix and eigenvalues to
        # largest. If points are superimposed, i.e. if the largest eigenvalue
        # != 0, no need to keep going
        # TODO: get rid of casting, this is a complex LAPACK issue currently
        w, v = np.linalg.eigh(A.astype(np.float64))
        w, v = w.astype(x.dtype), v.astype(x.dtype)
        if w[2] == 0.0:
            feats[k] = np.concatenate((center, np.zeros(12), np.array([len(clust)])))
            continue
        dirwt = 1.0 - w[1] / w[2]
        B = A / w[2]

        # Get the principal direction, identify the direction of the spread
        v0 = v[:, 2]

        # Projection all points, x, along the principal axis
        x0 = x.dot(v0)

        # Evaluate the distance from the points to the principal axis
        xp0 = x - np.outer(x0, v0)
        np0 = np.empty(len(xp0), dtype=data.dtype)
        for i in range(len(xp0)):
            np0[i] = np.linalg.norm(xp0[i])

        # Flip the principal direction if it is not pointing towards the
        # maximum spread
        sc = np.dot(x0, np0)
        if sc < 0:
            # Numba does not support unary `-`, have to flip manually
            v0 = np.zeros(3, dtype=data.dtype) - v0

        # Weight direction
        v0 = dirwt * v0

        # Append
        feats[k] = np.concatenate((voxel, B.flatten(), v0, np.array([len(x)])))

    return np.vstack(feats)



"""Module to convert a point cloud into an ellipsoidal envelope."""

import time

import numpy as np
import plotly.graph_objs as go
from scipy.special import gammaincinv  # pylint: disable=E0611

__all__ = ["ellipsoid_trace", "ellipsoid_traces"]


def ellipsoid_trace(
    points=None,
    centroid=None,
    covmat=None,
    contour=0.5,
    num_samples=10,
    color=None,
    intensity=None,
    hovertext=None,
    showscale=False,
    **kwargs,
):
    """Converts a cloud of points or a covariance matrix into a 3D ellipsoid.

    This function uses the centroid and the covariance matrix of a cloud of
    points to define an ellipsoid which would encompass a user-defined fraction
    `contour` of all the points, were the points distributed following
    a 3D Gaussian.

    Parameters
    ----------
    points : np.ndarray, optional
        (N, 3) Array of point coordinates
    centroid : np.ndarray, optional
        (3) Centroid
    covmat : np.ndarray, optional
        (3, 3) Covariance matrix which defines the ellipsoid shape
    contour : float, default 0.5
        Fraction of the points contained in the ellipsoid, under the
        Gaussian distribution assumption
    num_samples : int, default 10
        Number of points sampled along theta and phi in the spherical coordinate
        system of the ellipsoid. A larger number increases the resolution.
    color : Union[str, float], optional
        Color of ellipse
    intensity : Union[int, float], optional
        Color intensity of the ellipsoid along the colorscale axis
    hovertext : Union[int, str, np.ndarray], optional
        Text associated with the ellipsoid
    showscale : bool, default False
        If True, show the colorscale of the :class:`plotly.graph_objs.Mesh3d`
    **kwargs : dict, optional
        Additional parameters to pass to the underlying
        :class:`plotly.graph_objs.Mesh3d` object
    """
    # Ensure that either a cloud of points or a covariance matrix is provided
    assert (points is not None) ^ (
        centroid is not None and covmat is not None
    ), "Must provide either `points` or both `centroid` and `covmat`."

    # Compute the points on a unit sphere
    phi = np.linspace(0, 2 * np.pi, num=num_samples)
    theta = np.linspace(-np.pi / 2, np.pi / 2, num=num_samples)
    phi, theta = np.meshgrid(phi, theta)
    x = np.cos(theta) * np.sin(phi)
    y = np.cos(theta) * np.cos(phi)
    z = np.sin(theta)
    unit_points = np.vstack((x.flatten(), y.flatten(), z.flatten())).T

    # Get the centroid and the covariance matrix, if needed
    covmat_provided = True
    if covmat is None:
        assert points is not None, (
            "Points must be provided to compute covariance matrix "
            "if it is not provided explicitly."
        )
        covmat_provided = False

        if len(points) > 1:
            centroid = np.mean(points, axis=0)
            covmat = np.cov((points - centroid).T)
        else:
            centroid = points[0]
            covmat = np.zeros((3, 3))

    # Diagonalize the covariance matrix, get rotation matrix
    w, v = np.linalg.eigh(covmat)
    diag = np.diag(np.sqrt(w))
    rotmat = np.dot(diag, v.T)

    # Compute the radius corresponding to the contour probability and rotate
    # the points into the basis of the covariance matrix
    radius = 1.0
    if not covmat_provided:
        assert (
            contour > 0.0 and contour < 1.0
        ), "The `contour` parameter should be a probability."
        radius = np.sqrt(2 * gammaincinv(1.5, contour))

    ell_points = centroid + radius * np.dot(unit_points, rotmat)

    # Convert the color provided to a set of intensities, if needed
    if color is not None and not isinstance(color, str):
        assert intensity is None, "Must not provide both `color` and `intensity`."
        intensity = np.full(len(ell_points), color)
        color = None

    # Update hovertemplate style
    hovertemplate = "x: %{x}<br>y: %{y}<br>z: %{z}"
    if hovertext is not None:
        if not np.isscalar(hovertext):
            hovertemplate += "<br>%{text}"
        else:
            hovertemplate += f"<br>{hovertext}"
            hovertext = None

    # Append Mesh3d object
    return go.Mesh3d(
        x=ell_points[:, 0],
        y=ell_points[:, 1],
        z=ell_points[:, 2],
        color=color,
        intensity=intensity,
        alphahull=0,
        showscale=showscale,
        hovertemplate=hovertemplate,
        **kwargs,
    )


def ellipsoid_traces(
    centroids,
    covmat,
    color=None,
    hovertext=None,
    cmin=None,
    cmax=None,
    shared_legend=True,
    legendgroup=None,
    showlegend=True,
    name=None,
    **kwargs,
):
    """Function which produces a list of plotly traces of ellipsoids given a
    list of centroids and one covariance matrix in x, y and z.

    Parameters
    ----------
    centroids : np.ndarray
        (N, 3) Positions of each of the ellipsoid centroids
    covmat : np.ndarray
        (3, 3) Covariance matrix which defines any of the base ellipsoid shape
    color : Union[str, np.ndarray], optional
        Color of ellipsoids or list of color of ellispoids
    hovertext : Union[int, str, np.ndarray], optional
        Text associated with every ellipsoid or each ellipsoid
    cmin : float, optional
        Minimum value along the color scale
    cmax : float, optional
        Maximum value along the color scale
    shared_legend : bool, default True
        If True, the plotly legend of all ellipsoids is shared as one
    legendgroup : str, optional
        Legend group to be shared between all ellipsoids
    showlegend : bool, default `True`
        Whether to show legends on not
    name : str, optional
        Name of the trace(s)
    **kwargs : dict, optional
        List of additional arguments to pass to the underlying list of
        :class:`plotly.graph_objs.Mesh3D`

    Returns
    -------
    Union[List[plotly.graph_objs.Mesh3D]]
        Ellipsoid traces
    """
    # Check the parameters
    assert (
        color is None or np.isscalar(color) or len(color) == len(centroids)
    ), "Specify one color for all ellipsoids, or one color per ellipsoid."
    assert (
        hovertext is None or np.isscalar(hovertext) or len(hovertext) == len(centroids)
    ), "Specify one hovertext for all ellipsoids, or one hovertext per ellipsoid."

    # If one color is provided per ellipsoid, give an associated hovertext
    if hovertext is None and isinstance(color, (list, tuple, np.ndarray)):
        hovertext = [f"Value: {v:0.3f}" for v in color]

    # If cmin/cmax are not provided, must build them so that all ellipsoids
    # share the same colorscale range (not guaranteed otherwise)
    if color is not None and isinstance(color, (list, tuple, np.ndarray)):
        if len(color) > 0:
            if cmin is None:
                cmin = np.min(color)
            if cmax is None:
                cmax = np.max(color)

    # If the legend is to be shared, make sure there is a common legend group
    if shared_legend and legendgroup is None:
        legendgroup = "group_" + str(time.time())

    # Loop over the list of ellipsoid centroids
    traces = []
    col, hov = color, hovertext
    for i, centroid in enumerate(centroids):
        # Fetch the right color/hovertext combination
        if color is not None and isinstance(color, (list, tuple, np.ndarray)):
            col = color[i]
        if hovertext is not None and isinstance(hovertext, (list, tuple, np.ndarray)):
            hov = hovertext[i]

        # If the legend is shared, only draw the legend of the first trace
        if shared_legend:
            showlegend = showlegend and i == 0
            name_i = name
        else:
            name_i = f"{name} {i}"

        # Append list of traces
        traces.append(
            ellipsoid_trace(
                centroid=centroid,
                covmat=covmat,
                color=col,
                hovertext=hov,
                cmin=cmin,
                cmax=cmax,
                legendgroup=legendgroup,
                showlegend=showlegend,
                name=name_i,
                **kwargs,
            )
        )

    return traces



"""Machine learning models for neutrino physics reconstruction.

This module handles the construction, training, and execution of deep learning models
for liquid argon time projection chamber (LArTPC) data analysis.

**Model Management:**
- `ModelManager`: Central class for model lifecycle management
  - Model instantiation and configuration
  - Training loop orchestration
  - Validation and testing
  - Checkpoint saving and loading
  - Multi-GPU support

**Supported Model Architectures:**
- **UResNet**: U-Net with ResNet blocks for semantic segmentation
- **PPN**: Point Proposal Network for particle endpoint detection
- **SPICE**: Spatial Point cloud Instance Clustering and Embedding
- **GrapPA**: Graph neural networks for particle analysis
- **Chain Models**: End-to-end reconstruction pipelines

**Key Features:**
- Modular architecture with configurable components
- Support for sparse and dense convolutions
- Graph neural network implementations
- Multi-task learning capabilities
- Automatic mixed precision training
- Distributed training support
- Comprehensive logging and monitoring

**Training Pipeline:**
1. Data loading and preprocessing
2. Model architecture selection
3. Loss function configuration
4. Optimizer and scheduler setup
5. Training loop with validation
6. Model evaluation and metrics
7. Result visualization and analysis

**Example Usage:**
```python
from spine.model import ModelManager

# Initialize model manager with configuration
manager = ModelManager(config)

# Train the model
manager.train(train_loader, val_loader)

# Run inference
results = manager.forward(data_batch)
```

The module integrates with the broader SPINE ecosystem for data I/O,
visualization, and post-processing workflows.
"""

from .manager import ModelManager



"""Module with a data class objects which represent output particles."""

from dataclasses import dataclass, field
from typing import List

import numpy as np
from scipy.spatial.distance import cdist

from spine.data.particle import Particle
from spine.utils.docstring import inherit_docstring
from spine.utils.globals import (
    PID_LABELS,
    PID_MASSES,
    PID_TO_PDG,
    SHAPE_LABELS,
    SHOWR_SHP,
    TRACK_SHP,
)

from .base import RecoBase, TruthBase

__all__ = ["RecoParticle", "TruthParticle"]


@dataclass(eq=False)
class ParticleBase:
    """Base particle-specific information.

    Attributes
    ----------
    fragments : List[object]
        List of fragments that make up this particle
    fragment_ids : np.ndarray
        List of Fragment IDs that make up this particle
    num_fragments : int
        Number of fragments that make up this particle
    interaction_id : int
        Index of the interaction this particle belongs to
    shape : int
        Semantic type (shower (0), track (1), Michel (2), delta (3),
        low energy scatter (4)) of this particle
    pid : int
        Particle species (Photon (0), Electron (1), Muon (2), Charged Pion (3),
        Proton (4), Kaon (5)) of this particle
    chi2_pid : int
        Particle species as predicted by the chi2 template method (Muon (2),
        Charged Pion (3), Proton (4), Kaon (5)) of this particle
    chi2_per_pid : np.ndarray
        (P) Array of chi2 values associated with each particle class
    pdg_code : int
        PDG code corresponding to the PID number
    is_primary : bool
        Whether this particle was the first in the particle group
    length : float
        Length of the particle (only assigned to track objects)
    start_point : np.ndarray
        (3) Particle start point
    end_point : np.ndarray
        (3) Particle end point (only assigned to track objects)
    start_dir : np.ndarray
        (3) Particle direction w.r.t. the start point
    end_dir : np.ndarray
        (3) Particle direction w.r.t. the end point (only assigned
        to track objects)
    mass : float
        Rest mass of the particle in MeV/c^2
    ke : float
        Kinetic energy of the particle in MeV
    calo_ke : float
        Kinetic energy reconstructed from the energy depositions alone in MeV
    csda_ke : float
        Kinetic energy reconstructed from the particle range in MeV
    csda_ke_per_pid : np.ndarray
        (P) Same as `csda_ke` but for every available track PID hypothesis
    mcs_ke : float
        Kinetic energy reconstructed using the MCS method in MeV
    mcs_ke_per_pid : np.ndarray
        (P) Same as `mcs_ke` but for every available track PID hypothesis
    momentum : np.ndarray
        3-momentum of the particle at the production point in MeV/c
    p : float
        Momentum magnitude of the particle at the production point in MeV/c
    is_crt_matched : bool
        True if the particle was matched to a CRT hit
    crt_ids : np.ndarray
        (C) Indices of the CRT hits the particle is matched to
    crt_times : np.ndarray
        (C) Times at which the CRT hits occurred in microseconds
    crt_scores : np.ndarray
        (C) Quality metric associated with the CRT matches
    is_valid : bool
        Whether this particle counts towards an interaction topology. This
        may be False if a particle is below some defined energy threshold.
    """

    fragments: List[object] = None
    fragment_ids: np.ndarray = None
    num_fragments: int = None
    interaction_id: int = -1
    shape: int = -1
    pid: int = -1
    chi2_pid: int = -1
    chi2_per_pid: np.ndarray = None
    pdg_code: int = -1
    is_primary: bool = False
    length: float = -1.0
    start_point: np.ndarray = None
    end_point: np.ndarray = None
    start_dir: np.ndarray = None
    end_dir: np.ndarray = None
    mass: float = -1.0
    ke: float = -1.0
    calo_ke: float = -1.0
    csda_ke: float = -1.0
    csda_ke_per_pid: np.ndarray = None
    mcs_ke: float = -1.0
    mcs_ke_per_pid: np.ndarray = None
    momentum: np.ndarray = None
    p: float = None
    is_crt_matched: bool = False
    crt_ids: np.ndarray = None
    crt_times: np.ndarray = None
    crt_scores: np.ndarray = None
    is_valid: bool = True

    # Fixed-length attributes
    _fixed_length_attrs = (
        ("start_point", 3),
        ("end_point", 3),
        ("start_dir", 3),
        ("end_dir", 3),
        ("momentum", 3),
        ("chi2_per_pid", len(PID_LABELS) - 1),
        ("csda_ke_per_pid", len(PID_LABELS) - 1),
        ("mcs_ke_per_pid", len(PID_LABELS) - 1),
    )

    # Variable-length attributes as (key, dtype) pairs
    _var_length_attrs = (
        ("fragments", object),
        ("fragment_ids", np.int32),
        ("crt_ids", np.int32),
        ("crt_times", np.float32),
        ("crt_scores", np.float32),
    )

    # Attributes specifying coordinates
    _pos_attrs = ("start_point", "end_point")

    # Attributes specifying vector components
    _vec_attrs = ("start_dir", "end_dir", "momentum")

    # Boolean attributes
    _bool_attrs = ("is_primary", "is_crt_matched", "is_valid")

    # Enumerated attributes
    _enum_attrs = (
        ("shape", tuple((v, k) for k, v in SHAPE_LABELS.items())),
        ("pid", tuple((v, k) for k, v in PID_LABELS.items())),
    )

    # Attributes that must never be stored to file
    _skip_attrs = ("fragments", "ppn_points")

    def __str__(self):
        """Human-readable string representation of the particle object.

        Results
        -------
        str
            Basic information about the particle properties
        """
        pid_label = PID_LABELS[self.pid]
        match = self.match_ids[0] if len(self.match_ids) > 0 else -1
        return (
            f"Particle(ID: {self.id:<3} | PID: {pid_label:<8} "
            f"| Primary: {self.is_primary:<2} "
            f"| Size: {self.size:<5} | Match: {match:<3})"
        )

    def reset_crt_match(self, typed=True):
        """Reset all the CRT hit matching attributes.

        Parameters
        ----------
        typed : bool, default True
            If `True`, the underlying arrays are reset to typed empty arrays
        """
        self.is_crt_matched = False
        if typed:
            self.crt_ids = np.empty(0, dtype=np.int32)
            self.crt_times = np.empty(0, dtype=np.float32)
            self.crt_scores = np.empty(0, dtype=np.float32)
        else:
            self.crt_ids = []
            self.crt_times = []
            self.crt_scores = []

    @property
    def num_fragments(self):
        """Number of fragments that make up this particle.

        Returns
        -------
        int
            Number of fragments that make up the particle instance
        """
        return len(self.fragment_ids)

    @num_fragments.setter
    def num_fragments(self, num_fragments):
        pass

    @property
    def pdg_code(self):
        """Translates the enumerated particle type to a sign-less PDG code.

        Returns
        -------
        int
            Reconstructed sign-less PDG code
        """
        return PID_TO_PDG[self.pid]

    @pdg_code.setter
    def pdg_code(self, pdg_code):
        pass

    @property
    def p(self):
        """Computes the magnitude of the initial momentum.

        Returns
        -------
        float
            Norm of the initial momentum vector
        """
        return np.linalg.norm(self.momentum)

    @p.setter
    def p(self, p):
        pass


@dataclass(eq=False)
@inherit_docstring(RecoBase, ParticleBase)
class RecoParticle(ParticleBase, RecoBase):
    """Reconstructed particle information.

    Attributes
    ----------
    pid_scores : np.ndarray
        (P) Array of softmax scores associated with each particle class
    primary_scores : np.ndarray
        (2) Array of softmax scores associated with secondary and primary
    ppn_ids : np.ndarray
        (M) List of indexes of PPN points associated with this particle
    ppn_points : np.ndarray
        (M, 3) List of PPN points tagged to this particle
    vertex_distance : float
        Set-to-point distance between all particle points and the parent
        interaction vertex position in cm
    start_dedx : float
        dE/dx around a user-defined neighborhood of the start point in MeV/cm
    end_dedx : float
        dE/dx around a user-defined neighborhood of the end point in MeV/cm
    start_straightness : float
        Explained variance ratio of the beginning of the particle
    directional_spread : float
        Estimate of the angular spread of the particle (cosine spread)
    axial_spread : float
        Pearson correlation coefficient of the axial profile of the particle
        w.r.t. to the distance from its start point
    """

    pid_scores: np.ndarray = None
    primary_scores: np.ndarray = None
    ppn_ids: np.ndarray = None
    ppn_points: np.ndarray = None
    vertex_distance: float = -1.0
    start_dedx: float = -1.0
    end_dedx: float = -1.0
    start_straightness: float = -1.0
    directional_spread: float = -1.0
    axial_spread: float = -np.inf

    # Fixed-length attributes
    _fixed_length_attrs = (
        ("pid_scores", len(PID_LABELS) - 1),
        ("primary_scores", 2),
        *ParticleBase._fixed_length_attrs,
    )

    # Variable-length attributes
    _var_length_attrs = (
        ("ppn_ids", np.int32),
        ("ppn_points", (3, np.float32)),
        *RecoBase._var_length_attrs,
        *ParticleBase._var_length_attrs,
    )

    # Boolean attributes
    _bool_attrs = (*RecoBase._bool_attrs, *ParticleBase._bool_attrs)

    # Attributes that must never be stored to file
    _skip_attrs = ("ppn_points", *RecoBase._skip_attrs, *ParticleBase._skip_attrs)

    def __str__(self):
        """Human-readable string representation of the particle object.

        Results
        -------
        str
            Basic information about the particle properties
        """
        return "Reco" + super().__str__()

    def merge(self, other):
        """Merge another particle instance into this one.

        The merging strategy differs depending on the the particle shapes
        merged together. There are two categories:
        - Track + track
          - The start/end points are produced by finding the combination of points
            which are farthest away from each other (one from each constituent)
          - The primary scores/primary status match that of the constituent
            particle with the highest primary score
          - The PID scores/PID value match that of the constituent particle with
            the highest primary score
        - Shower + Track
          - The track is always merged into the shower, not the other way around
          - The start point of the shower is updated to be the track end point
          further away from the current shower start point
          - The primary scores/primary status match that of the constituent
            particle with the highest primary score
          - The PID scores/PID value is kept unchanged (that of the shower)

        Parameters
        ----------
        other : RecoParticle
            Other reconstructed particle to merge into this one
        """
        # Check that the particles being merged fit one of two categories
        assert (
            self.shape in (SHOWR_SHP, TRACK_SHP) and other.shape == TRACK_SHP
        ), "Can only merge two track particles or a track into a shower."

        # Check that neither particle has yet been matched
        assert (
            not self.is_matched and not other.is_matched
        ), "Cannot merge particles that already have matches."

        # Concatenate the two particle long-form attributes together
        for attr in self._cat_attrs:
            val = np.concatenate([getattr(self, attr), getattr(other, attr)])
            setattr(self, attr, val)

        # Select end points and end directions appropriately
        if self.shape == TRACK_SHP:
            # If two tracks, pick points furthest apart
            points_i = np.vstack([self.start_point, self.end_point])
            points_j = np.vstack([other.start_point, other.end_point])
            dirs_i = np.vstack([self.start_dir, self.end_dir])
            dirs_j = np.vstack([other.start_dir, other.end_dir])

            dists = cdist(points_i, points_j)
            max_index = np.argmax(dists)
            max_i, max_j = max_index // 2, max_index % 2

            self.start_point = points_i[max_i]
            self.end_point = points_j[max_j]
            self.start_dir = dirs_i[max_i]
            self.end_dir = dirs_j[max_j]

        else:
            # If a shower and a track, pick track point furthest from shower
            points_i = self.start_point.reshape(-1, 3)
            points_j = np.vstack([other.start_point, other.end_point])
            dirs_j = np.vstack([other.start_dir, other.end_dir])

            dists = cdist(points_i, points_j)
            max_j = np.argmax(dists)

            self.start_point = points_j[max_j]
            self.start_dir = dirs_j[max_j]

        # Match primary/PID to the most primary particle
        if other.primary_scores[-1] > self.primary_scores[-1]:
            self.primary_scores = other.primary_scores
            self.is_primary = other.is_primary
            if self.shape == TRACK_SHP:
                self.pid_scores = other.pid_scores
                self.pid = other.pid

        # If the calorimetric KEs have been computed, can safely sum
        if other.calo_ke > 0.0:
            self.calo_ke += other.calo_ke

    @property
    def mass(self):
        """Rest mass of the particle in MeV/c^2.

        The mass is inferred from the predicted mass.

        Returns
        -------
        float
            Rest mass of the particle
        """
        if self.pid in PID_MASSES:
            return PID_MASSES[self.pid]

        return -1.0

    @mass.setter
    def mass(self, mass):
        pass

    @property
    def ke(self):
        """Best-guess kinetic energy in MeV.

        Uses calorimetry for EM activity and this order for track:
        - CSDA-based estimate if it is available
        - MCS-based estimate if it is available
        - Calorimetry if all else fails

        Returns
        -------
        float
            Best-guess kinetic energy
        """
        if self.shape != TRACK_SHP:
            # If a particle is not a track, can only use calorimetry
            return self.calo_ke

        else:
            # If a particle is a track, pick CSDA for contained tracks and
            # pick MCS for uncontained tracks, unless specified otherwise
            if self.is_contained and self.csda_ke > 0.0:
                return self.csda_ke
            elif not self.is_contained and self.mcs_ke > 0.0:
                return self.mcs_ke
            else:
                return self.calo_ke

    @ke.setter
    def ke(self, ke):
        pass

    @property
    def momentum(self):
        """Best-guess momentum in MeV/c.

        Returns
        -------
        np.ndarray
            (3) Momentum vector
        """
        ke = self.ke
        if ke >= 0.0 and self.start_dir[0] != -np.inf and self.pid in PID_MASSES:
            mass = PID_MASSES[self.pid]
            mom = np.sqrt(ke**2 + 2 * ke * mass)
            return mom * self.start_dir

        else:
            return np.full(3, -np.inf, dtype=np.float32)

    @momentum.setter
    def momentum(self, momentum):
        pass

    @property
    def reco_ke(self):
        """Alias for `ke`, to match nomenclature in truth."""
        return self.ke

    @property
    def reco_momentum(self):
        """Alias for `momentum`, to match nomenclature in truth."""
        return self.momentum

    @property
    def reco_length(self):
        """Alias for `length`, to match nomenclature in truth."""
        return self.length

    @property
    def reco_start_dir(self):
        """Alias for `start_dir`, to match nomenclature in truth."""
        return self.start_dir

    @property
    def reco_end_dir(self):
        """Alias for `end_dir`, to match nomenclature in truth."""
        return self.end_dir


@dataclass(eq=False)
@inherit_docstring(TruthBase, ParticleBase)
class TruthParticle(Particle, ParticleBase, TruthBase):
    """Truth particle information.

    This inherits all of the attributes of :class:`Particle`, which contains
    the G4 truth information for the particle.

    Attributes
    ----------
    orig_interaction_id : int
        Unaltered index of the interaction in the original MC particle list
    orig_parent_id : int
        Unaltered index of the particle parent in the original MC particle list
    orig_group_id : int
        Unaltered index of the particle group in the original MC particle list
    orig_children_id : np.ndarray
        Unaltered list of the particle children in the original MC particle list
    children_counts : np.ndarray
        (P) Number of truth child particle of each shape
    reco_length : float
        Reconstructed length of the particle (only assigned to track objects)
    reco_start_dir : np.ndarray
        (3) Particle direction estimate w.r.t. the start point
    reco_end_dir : np.ndarray
        (3) Particle direction estimate w.r.t. the end point (only assigned
        to track objects)
    reco_ke : float
        Best-guess reconstructed KE of the particle
    reco_momentum : np.ndarray
        Best-guess reconstructed momentum of the particle
    """

    orig_interaction_id: int = -1
    orig_parent_id: int = -1
    orig_group_id: int = -1
    orig_children_id: np.ndarray = None
    children_counts: np.ndarray = None
    reco_length: float = -1.0
    reco_start_dir: np.ndarray = None
    reco_end_dir: np.ndarray = None
    reco_ke: float = -1.0
    reco_momentum: np.ndarray = None

    # Fixed-length attributes
    _fixed_length_attrs = (
        ("reco_start_dir", 3),
        ("reco_end_dir", 3),
        ("reco_momentum", 3),
        *ParticleBase._fixed_length_attrs,
        *Particle._fixed_length_attrs,
    )

    # Variable-length attributes
    _var_length_attrs = (
        ("orig_children_id", np.int64),
        ("children_counts", np.int32),
        *TruthBase._var_length_attrs,
        *ParticleBase._var_length_attrs,
        *Particle._var_length_attrs,
    )

    # Attributes specifying coordinates
    _pos_attrs = (*ParticleBase._pos_attrs, *Particle._pos_attrs)

    # Attributes specifying vector components
    _vec_attrs = (
        "reco_start_dir",
        "reco_end_dir",
        "reco_momentum",
        *ParticleBase._vec_attrs,
        *Particle._vec_attrs,
    )

    # Boolean attributes
    _bool_attrs = (*TruthBase._bool_attrs, *ParticleBase._bool_attrs)

    # Attributes that must never be stored to file
    _skip_attrs = (*TruthBase._skip_attrs, *ParticleBase._skip_attrs)

    def __str__(self):
        """Human-readable string representation of the particle object.

        Results
        -------
        str
            Basic information about the particle properties
        """
        return "Truth" + super().__str__()

    @property
    def start_dir(self):
        """Converts the initial momentum to a direction vector.

        Returns
        -------
        np.ndarray
            (3) Start direction vector
        """
        if self.momentum is not None:
            norm = np.linalg.norm(self.momentum)
            if norm > 0.0 and norm != np.inf:
                return self.momentum / norm

        return np.full(3, -np.inf, dtype=np.float32)

    @start_dir.setter
    def start_dir(self, start_dir):
        pass

    @property
    def end_dir(self):
        """Converts the final momentum to a direction vector.

        Note that if a particle stops, this is unreliable as an estimate of the
        direction of the particle before it stops.

        Returns
        -------
        np.ndarray
            (3) End direction vector
        """
        if self.end_momentum is not None:
            norm = np.linalg.norm(self.end_momentum)
            if self.shape == TRACK_SHP and norm > 0.0 and norm != np.inf:
                return self.end_momentum / norm

        return np.full(3, -np.inf, dtype=np.float32)

    @end_dir.setter
    def end_dir(self, end_dir):
        pass

    @property
    def ke(self):
        """Converts the particle initial energy to a kinetic energy.

        This only works for particles with a known mass (as defined in
        `spine.utils.globals`).

        Returns
        -------
        float
            Initial kinetic energy of the particle
        """
        if self.mass > -1.0:
            return self.energy_init - self.mass

        return -1.0

    @ke.setter
    def ke(self, ke):
        pass

    @property
    def reco_ke(self):
        """Best-guess reconstructed kinetic energy in MeV.

        Uses calorimetry for EM activity and this order for track:
        - CSDA-based estimate if it is available
        - MCS-based estimate if it is available
        - Calorimetry if all else fails

        Returns
        -------
        float
            Best-guess kinetic energy
        """
        if self.shape != TRACK_SHP:
            # If a particle is not a track, can only use calorimetry
            return self.calo_ke

        else:
            # If a particle is a track, pick CSDA for contained tracks and
            # pick MCS for uncontained tracks, unless specified otherwise
            if self.is_contained and self.csda_ke > 0.0:
                return self.csda_ke
            elif not self.is_contained and self.mcs_ke > 0.0:
                return self.mcs_ke
            else:
                return self.calo_ke

    @reco_ke.setter
    def reco_ke(self, reco_ke):
        pass

    @property
    def reco_momentum(self):
        """Best-guess reconstructed momentum in MeV/c.

        Returns
        -------
        np.ndarray
            (3) Momentum vector
        """
        ke = self.reco_ke
        if ke >= 0.0 and self.reco_start_dir[0] != -np.inf and self.pid in PID_MASSES:
            mass = PID_MASSES[self.pid]
            mom = np.sqrt(ke**2 + 2 * ke * mass)
            return mom * self.reco_start_dir

        else:
            return np.full(3, -np.inf, dtype=np.float32)

    @reco_momentum.setter
    def reco_momentum(self, reco_momentum):
        pass



"""Module with a data class object which represents optical information.

This copies the internal structure of :class:`larcv.Flash`.
"""

from dataclasses import dataclass

import numpy as np

from .base import PosDataBase

__all__ = ["Flash"]


@dataclass(eq=False)
class Flash(PosDataBase):
    """Optical flash information.

    Attributes
    ----------
    id : int
        Index of the flash in the list
    volume_id : int
        Index of the optical volume in which the flash was recorded
    time : float
        Time with respect to the trigger in microseconds
    time_width : float
        Width of the flash in microseconds
    time_abs : float
        Time in units of PMT readout clock
    frame : int
        Frame number
    in_beam_frame : bool
        Whether the flash is in the beam frame
    on_beam_time : bool
        Whether the flash time is consistent with the beam window
    total_pe : float
        Total number of PE in the flash
    fast_to_total : float
        Fraction of the total PE contributed by the fast component
    pe_per_ch : np.ndarray
        (N) Fixed-length array of the number of PE per optical detector
    center : np.ndarray
        Barycenter of the flash in detector coordinates
    width : np.ndarray
        Spatial width of the flash in detector coordinates
    units : str
        Units in which the position coordinates are expressed
    """

    id: int = -1
    volume_id: int = -1
    frame: int = -1
    in_beam_frame: bool = False
    on_beam_time: bool = False
    time: float = -1.0
    time_width: float = -1.0
    time_abs: float = -1.0
    total_pe: float = -1.0
    fast_to_total: float = -1.0
    pe_per_ch: np.ndarray = None
    center: np.ndarray = None
    width: np.ndarray = None
    units: str = "cm"

    # Fixed-length attributes
    _fixed_length_attrs = (("center", 3), ("width", 3))

    # Variable-length attributes
    _var_length_attrs = (("pe_per_ch", np.float32),)

    # Attributes specifying coordinates
    _pos_attrs = ("center",)

    # Attributes specifying vector components
    _vec_attrs = ("width",)

    # Index attributes
    _index_attrs = ("id",)

    @classmethod
    def from_larcv(cls, flash):
        """Builds and returns a Flash object from a LArCV Flash object.

        Parameters
        ----------
        flash : larcv.Flash
            LArCV-format optical flash

        Returns
        -------
        Flash
            Flash object
        """
        # Get the physical center and width of the flash
        axes = ("x", "y", "z")
        center = np.array([getattr(flash, f"{a}Center")() for a in axes])
        width = np.array([getattr(flash, f"{a}Width")() for a in axes])

        # Get the number of PEs per optical channel
        pe_per_ch = np.array(list(flash.PEPerOpDet()), dtype=np.float32)

        # Get the volume ID, if it is filled (TODO: simplify with update)
        volume_id = -1
        for attr in ("tpc", "volume_id"):
            if hasattr(flash, attr):
                volume_id = getattr(flash, attr)()

        return cls(
            id=flash.id(),
            volume_id=volume_id,
            frame=flash.frame(),
            in_beam_frame=flash.inBeamFrame(),
            on_beam_time=flash.onBeamTime(),
            time=flash.time(),
            time_abs=flash.absTime(),
            time_width=flash.timeWidth(),
            total_pe=flash.TotalPE(),
            pe_per_ch=pe_per_ch,
            center=center,
            width=width,
        )

    def merge(self, other):
        """Merge another flash into this one.

        The merging strategy proceeds as follows:
        - The earlier flash takes precedence over the later flash as far as
          all timing-related information is concerned (flash time, etc.)
        - The combined flash centroid is produced by taking the weighted average
          of the two existing flash centroids
        - The PE values in each light collection system are added together, so
          is the total PE value of the combined flash

        Parameters
        ----------
        other : Flash
            Flash to merge into this one
        """
        # Check that the position units are the same
        assert (
            self.units == other.units
        ), "The units of the flash to be merged do not match."

        # Determine the flash window end points (to merge time widths later)
        end_i, end_j = self.time + self.time_width, other.time + other.time_width

        # If the other flash happened first, update the timing information
        if self.time > other.time:
            self.time = other.time
            self.time_abs = other.time_abs
            self.on_beam_time = other.on_beam_time
            self.frame = other.frame
            self.in_beam_frame = other.in_beam_frame

        # Take the union of the two time widths as the new combined width
        self.time_width = max(end_i, end_j) - self.time

        # Take the weighted average of the centroids to compute the new one
        valid_mask = (self.width > 0.0) & (other.width > 0.0)

        w_i, w_j = 1.0 / self.width**2, 1.0 / other.width**2
        self.center = (w_i * self.center + w_j * other.center) / (w_i + w_j)

        self.width = 1.0 / np.sqrt(w_i + w_j)
        self.width[~valid_mask] = -1.0

        # Compute the new total PE and fast light component to total ratio
        t_i, t_j = self.total_pe, other.total_pe
        self.total_pe = t_i + t_j

        r_i, r_j = self.fast_to_total, other.fast_to_total
        self.fast_to_total = (r_i * t_i + r_j * t_j) / (t_i + t_j)

        # Merge the PE count in each PMT
        pe_per_ch = np.zeros(
            max(len(self.pe_per_ch), len(other.pe_per_ch)), dtype=self.pe_per_ch.dtype
        )
        pe_per_ch[: len(self.pe_per_ch)] += self.pe_per_ch
        pe_per_ch[: len(other.pe_per_ch)] += other.pe_per_ch

        self.pe_per_ch = pe_per_ch

        # The new volume ID is invalid if the two original volumes differ
        if self.volume_id != other.volume_id:
            self.volume_id = -1



"""Module with a data class object which represents the run information.

It can extract run attributes from any event-level LArCV object.
"""

from dataclasses import dataclass

from .base import DataBase

__all__ = ["RunInfo"]


@dataclass(eq=False)
class RunInfo(DataBase):
    """Run information related to a specific event.

    Attributes
    ----------
    run : int
        Run ID
    subrun : int
        Sub-run ID
    event : int
        Event ID
    """

    run: int = -1
    subrun: int = -1
    event: int = -1

    @classmethod
    def from_larcv(cls, larcv_event):
        """
        Builds and returns a Meta object from a LArCV 2D metadata object

        Parameters
        ----------
        larcv_event : larcv.EventBase
             LArCV event object which contains the run information as attributes

        Returns
        -------
        Meta
            Metadata object
        """
        return cls(
            run=larcv_event.run(),
            subrun=larcv_event.subrun(),
            event=larcv_event.event(),
        )



"""Particle direction reconstruction module."""

import numpy as np

from spine.post.base import PostBase
from spine.utils.globals import TRACK_SHP
from spine.utils.gnn.cluster import get_cluster_directions

__all__ = ["DirectionProcessor"]


class DirectionProcessor(PostBase):
    """Reconstructs the direction of fragments and/or particles w.r.t. to
    their start (and end for tracks) points.

    This modules assign the `start_dir` and `end_dir` attributes.
    """

    # Name of the post-processor (as specified in the configuration)
    name = "direction"

    # Alternative allowed names of the post-processor
    aliases = ("reconstruct_directions",)

    def __init__(
        self,
        radius=-1,
        optimize=True,
        obj_type="particle",
        truth_point_mode="points",
        run_mode="both",
    ):
        """Store the particle direction reconstruction parameters.

        Parameters
        ----------
        radius : float, default -1
            Radius around the start voxel to include in the direction estimate
        optimize : bool, default True
            Optimize the number of points involved in the direction estimate
        """
        # Initialize the parent class
        super().__init__(obj_type, run_mode, truth_point_mode)

        # Store the direction reconstruction parameters
        self.radius = radius
        self.optimize = optimize

        # Make sure the voxel coordinates are provided as a tensor
        if run_mode != "truth":
            self.update_keys({"points": True})
        if run_mode != "reco":
            self.update_keys({self.truth_point_key: True})

    def process(self, data):
        """Reconstruct the directions of all particles in one entry.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # Loop over particle objects
        for k in self.obj_keys:
            # In order to parallelize the process, fetch all start points and
            # clusters once before passing them to the a parallelized function
            points_key = "points" if not "truth" in k else self.truth_point_key
            attrs, part_ids = [], []
            clusts, ref_points = [], []
            for obj in data[k]:
                # Get index
                index = self.get_index(obj)
                if len(index) == 0:
                    continue

                # Store index, start point and the relevant mapping
                for side in ("start", "end"):
                    if side == "end" and obj.shape != TRACK_SHP:
                        continue

                    attr = f"reco_{side}_dir" if obj.is_truth else f"{side}_dir"
                    attrs.append(attr)
                    part_ids.append(obj.id)

                    clusts.append(index)
                    ref_points.append(getattr(obj, f"{side}_point"))

            # Check that there is at least one direction to compute
            if len(clusts) < 1:
                continue

            # Compute the direction vectors
            dirs = get_cluster_directions(
                data[points_key],
                np.vstack(ref_points),
                clusts,
                max_dist=self.radius,
                optimize=self.optimize,
            )

            # Assign directions to the appropriate particles
            for i, part_id in enumerate(part_ids):
                if attrs[i].startswith("start"):
                    setattr(data[k][part_id], attrs[i], dirs[i])
                else:
                    setattr(data[k][part_id], attrs[i], -dirs[i])



from collections import defaultdict

import torch
import torch.nn as nn

from .lovasz import StableBCELoss, iou_binary, lovasz_hinge_flat
from .misc import *


class MaskBCELoss(nn.Module):
    """
    Loss function for Sparse Spatial Embeddings Model, with fixed
    centroids and symmetric gaussian kernels.
    """

    def __init__(self, cfg, name="spice_loss"):
        super(MaskBCELoss, self).__init__()
        self.loss_config = cfg.get(name, {})
        self.seediness_weight = self.loss_config.get("seediness_weight", 0.0)
        self.embedding_weight = self.loss_config.get("embedding_weight", 1.0)
        self.smoothing_weight = self.loss_config.get("smoothing_weight", 1.0)
        self.spatial_size = self.loss_config.get("spatial_size", 512)
        self.loss_scheme = self.loss_config.get("loss_scheme", "BCE")

        if self.loss_scheme == "BCE":
            self.mask_loss = StableBCELoss()
        elif self.loss_scheme == "lovasz_hinge":
            self.mask_loss = lovasz_hinge_flat
        elif self.loss_scheme == "focal":
            raise NotImplementedError
        else:
            raise ValueError("Invalid loss scheme: {}".format(self.loss_scheme))

        # BCELoss for Embedding Loss
        self.bceloss = StableBCELoss()
        # L2 Loss for Seediness and Smoothing
        self.l2loss = torch.nn.L1Loss(reduction="mean")

    def find_cluster_means(self, features, labels):
        """
        For a given image, compute the centroids mu_c for each
        cluster label in the embedding space.
        Inputs:
            features (torch.Tensor): the pixel embeddings, shape=(N, d) where
            N is the number of pixels and d is the embedding space dimension.
            labels (torch.Tensor): ground-truth group labels, shape=(N, )
        Returns:
            cluster_means (torch.Tensor): (n_c, d) tensor where n_c is the number of
            distinct instances. Each row is a (1,d) vector corresponding to
            the coordinates of the i-th centroid.
        """
        clabels = labels.unique(sorted=True)
        cluster_means = []
        for c in clabels:
            index = labels == c
            mu_c = features[index].mean(0)
            cluster_means.append(mu_c)
        cluster_means = torch.stack(cluster_means)
        return cluster_means

    def get_per_class_probabilities(self, embeddings, margins, labels, coords):
        """
        Computes binary foreground/background loss.
        """
        device = embeddings.device
        loss = 0.0
        smoothing_loss = 0.0
        centroids = self.find_cluster_means(coords, labels)
        n_clusters = len(centroids)
        cluster_labels = labels.unique(sorted=True)
        probs = torch.zeros(embeddings.shape[0]).float().to(device)
        acc = 0.0

        for i, c in enumerate(cluster_labels):
            index = labels == c
            mask = torch.zeros(embeddings.shape[0]).to(device)
            mask[index] = 1.0
            mask[~index] = 0.0
            sigma = torch.mean(margins[index], dim=0)
            dists = torch.sum(torch.pow(embeddings - centroids[i], 2), dim=1)
            p = torch.clamp(torch.exp(-dists / (2 * torch.pow(sigma, 2))), min=0, max=1)
            probs[index] = p[index]
            loss += self.bceloss(p, mask)
            acc += iou_binary(p > 0.5, mask)
            sigma_detach = sigma.detach()
            smoothing_loss += torch.sum(torch.pow(margins[index] - sigma_detach, 2))

        loss /= n_clusters
        smoothing_loss /= n_clusters
        acc /= n_clusters

        return loss, smoothing_loss, probs, acc

    def combine_multiclass(
        self, embeddings, margins, seediness, slabels, clabels, coords
    ):
        """
        Wrapper function for combining different components of the loss,
        in particular when clustering must be done PER SEMANTIC CLASS.

        NOTE: When there are multiple semantic classes, we compute the DLoss
        by first masking out by each semantic segmentation (ground-truth/prediction)
        and then compute the clustering loss over each masked point cloud.

        INPUTS:
            features (torch.Tensor): pixel embeddings
            slabels (torch.Tensor): semantic labels
            clabels (torch.Tensor): group/instance/cluster labels

        OUTPUT:
            loss_segs (list): list of computed loss values for each semantic class.
            loss[i] = computed DLoss for semantic class <i>.
            acc_segs (list): list of computed clustering accuracy for each semantic class.
        """
        loss = defaultdict(list)
        accuracy = defaultdict(float)
        semantic_classes = slabels.unique()
        # print(semantic_classes)
        for sc in semantic_classes:
            if int(sc) == 4:
                continue
            index = slabels == sc
            mask_loss, smoothing_loss, probs, acc = self.get_per_class_probabilities(
                embeddings[index], margins[index], clabels[index], coords[index]
            )
            prob_truth = probs.detach()
            seed_loss = self.l2loss(prob_truth, seediness[index].squeeze(1))
            total_loss = (
                self.embedding_weight * mask_loss
                + self.seediness_weight * seed_loss
                + self.smoothing_weight * smoothing_loss
            )
            loss["loss"].append(total_loss)
            loss["mask_loss"].append(float(self.embedding_weight * mask_loss))
            loss["seed_loss"].append(float(self.seediness_weight * seed_loss))
            loss["smoothing_loss"].append(float(self.smoothing_weight * smoothing_loss))
            loss["mask_loss_{}".format(int(sc))].append(float(mask_loss))
            loss["seed_loss_{}".format(int(sc))].append(float(seed_loss))
            accuracy["accuracy_{}".format(int(sc))] = acc

        return loss, accuracy

    def forward(self, out, segment_label, group_label):

        num_gpus = len(segment_label)
        loss = defaultdict(list)
        accuracy = defaultdict(list)

        for i in range(num_gpus):
            slabels = segment_label[i][:, -1]
            # coords = segment_label[i][:, :3].float()
            # if torch.cuda.is_available():
            #    coords = coords.cuda()
            slabels = slabels.int()
            clabels = group_label[i][:, -1]
            batch_idx = segment_label[i][:, 3]
            embedding = out["embeddings"][i]
            seediness = out["seediness"][i]
            margins = out["margins"][i]
            nbatch = batch_idx.unique().shape[0]

            for bidx in batch_idx.unique(sorted=True):
                embedding_batch = embedding[batch_idx == bidx]
                slabels_batch = slabels[batch_idx == bidx]
                clabels_batch = clabels[batch_idx == bidx]
                seed_batch = seediness[batch_idx == bidx]
                margins_batch = margins[batch_idx == bidx]

                loss_class, acc_class = self.combine_multiclass(
                    embedding_batch,
                    margins_batch,
                    seed_batch,
                    slabels_batch,
                    clabels_batch,
                )
                for key, val in loss_class.items():
                    loss[key].append(sum(val) / len(val))
                for s, acc in acc_class.items():
                    accuracy[s].append(acc)
                acc = sum(acc_class.values()) / len(acc_class.values())
                accuracy["accuracy"].append(acc)

        loss_avg = {}
        acc_avg = defaultdict(float)

        for key, val in loss.items():
            loss_avg[key] = sum(val) / len(val)
        for key, val in accuracy.items():
            acc_avg[key] = sum(val) / len(val)

        res = {}
        res.update(loss_avg)
        res.update(acc_avg)

        return res


class MaskBCELoss2(MaskBCELoss):
    """
    Spatial Embeddings Loss with trainable center of attention.
    """

    def __init__(self, cfg, name="spice_loss"):
        super(MaskBCELoss2, self).__init__(cfg, name)

    def get_per_class_probabilities(self, embeddings, margins, labels, coords):
        """
        Computes binary foreground/background loss.
        """
        device = embeddings.device
        loss = 0.0
        smoothing_loss = 0.0
        centroids = self.find_cluster_means(embeddings, labels)
        n_clusters = len(centroids)
        cluster_labels = labels.unique(sorted=True)
        probs = torch.zeros(embeddings.shape[0]).float().to(device)
        acc = 0.0

        for i, c in enumerate(cluster_labels):
            index = labels == c
            mask = torch.zeros(embeddings.shape[0]).to(device)
            mask[index] = 1.0
            mask[~index] = 0.0
            sigma = torch.mean(margins[index], dim=0)
            dists = torch.sum(torch.pow(embeddings - centroids[i], 2), dim=1)
            p = torch.clamp(torch.exp(-dists / (2 * torch.pow(sigma, 2))), min=0, max=1)
            probs[index] = p[index]
            loss += self.bceloss(p, mask)
            acc += iou_binary(p > 0.5, mask, per_image=False)
            sigma_detach = sigma.detach()
            smoothing_loss += torch.sum(torch.pow(margins[index] - sigma_detach, 2))

        loss /= n_clusters
        smoothing_loss /= n_clusters
        acc /= n_clusters

        return loss, smoothing_loss, probs, acc


class MaskBCELossBivariate(MaskBCELoss):
    """
    Spatial Embeddings Loss with trainable center of attraction and
    bivariate gaussian probability kernels.
    """

    def __init__(self, cfg, name="spice_loss"):
        super(MaskBCELossBivariate, self).__init__(cfg, name)

    def get_per_class_probabilities(self, embeddings, margins, labels, coords):
        """
        Computes binary foreground/background loss.
        """
        device = embeddings.device
        loss = 0.0
        smoothing_loss = 0.0
        centroids = self.find_cluster_means(embeddings, labels)
        n_clusters = len(centroids)
        cluster_labels = labels.unique(sorted=True)
        probs = torch.zeros(embeddings.shape[0]).float().to(device)
        acc = 0.0

        for i, c in enumerate(cluster_labels):
            index = labels == c
            mask = torch.zeros(embeddings.shape[0]).to(device)
            mask[index] = 1.0
            mask[~index] = 0.0
            sigma = torch.mean(margins[index], dim=0)
            dists = torch.pow(embeddings - centroids[i], 2)
            dists = dists / (2 * torch.pow(sigma, 2))
            p = torch.clamp(torch.exp(-torch.sum(dists, dim=1)), min=0, max=1)
            probs[index] = p[index]
            loss += self.bceloss(p, mask)
            sigma_detach = sigma.detach()
            smoothing_loss += torch.sum(torch.pow(margins[index] - sigma_detach, 2))

        loss /= n_clusters
        smoothing_loss /= n_clusters
        acc /= n_clusters

        return loss, smoothing_loss, probs, acc


class MaskLovaszHingeLoss(MaskBCELoss2):
    """
    Spatial Embeddings Loss using Lovasz Hinge for foreground/background
    segmentation and trainable center of attention.
    """

    def __init__(self, cfg, name="spice_loss"):
        super(MaskLovaszHingeLoss, self).__init__(cfg, name)

    def get_per_class_probabilities(self, embeddings, margins, labels):
        """
        Computes binary foreground/background loss.
        """
        device = embeddings.device
        loss = 0.0
        smoothing_loss = 0.0
        centroids = self.find_cluster_means(embeddings, labels)
        n_clusters = len(centroids)
        cluster_labels = labels.unique(sorted=True)
        probs = torch.zeros(embeddings.shape[0]).float().to(device)
        acc = 0.0

        for i, c in enumerate(cluster_labels):
            index = labels == c
            mask = torch.zeros(embeddings.shape[0]).to(device)
            mask[index] = 1
            mask[~index] = 0
            sigma = torch.mean(margins[index], dim=0)
            dists = torch.sum(torch.pow(embeddings - centroids[i], 2), dim=1)
            p = torch.exp(-dists / (2 * torch.pow(sigma, 2) + 1e-6))
            probs[index] = p[index]
            loss += lovasz_hinge_flat(2 * p - 1, mask)
            sigma_detach = sigma.detach()
            smoothing_loss += torch.sum(torch.pow(margins[index] - sigma_detach, 2))

        loss /= n_clusters
        smoothing_loss /= n_clusters
        acc /= n_clusters

        return loss, smoothing_loss, probs, acc


class CELovaszLoss(MaskBCELoss2):

    def __init__(self, cfg, name="spice_loss"):
        super(CELovaszLoss, self).__init__(cfg, name)

    def get_per_class_probabilities(self, embeddings, margins, labels, coords):
        """
        Computes binary foreground/background loss.
        """
        device = embeddings.device
        loss = 0.0
        smoothing_loss = 0.0
        centroids = self.find_cluster_means(embeddings, labels)
        n_clusters = len(centroids)
        cluster_labels = labels.unique(sorted=True)
        probs = torch.zeros(embeddings.shape[0]).float().to(device)
        acc = 0.0

        for i, c in enumerate(cluster_labels):
            index = labels == c
            mask = torch.zeros(embeddings.shape[0]).to(device)
            mask[index] = 1.0
            mask[~index] = 0.0
            sigma = torch.mean(margins[index], dim=0)
            dists = torch.sum(torch.pow(embeddings - centroids[i], 2), dim=1)
            p = torch.clamp(
                torch.exp(-dists / (2 * torch.pow(sigma, 2) + 1e-8)), min=0, max=1
            )
            probs[index] = p[index]
            loss += (self.bceloss(p, mask) + lovasz_hinge_flat(2.0 * p - 1, mask)) / 2
            acc += iou_binary(p > 0.5, mask, per_image=False)
            sigma_detach = sigma.detach()
            smoothing_loss += torch.mean(
                torch.norm(margins[index] - sigma_detach, dim=1)
            )

        loss /= n_clusters
        smoothing_loss /= n_clusters
        acc /= n_clusters

        return loss, smoothing_loss, probs, acc


class MaskLovaszInterLoss(MaskLovaszHingeLoss):

    def __init__(self, cfg, name="spice_loss"):
        super(MaskLovaszInterLoss, self).__init__(cfg, name)
        self.inter_weight = self.loss_config.get("inter_weight", 1.0)
        self.norm = 2

    def regularization(self, cluster_means):
        """
        Implementation of regularization loss in Discriminative Loss
        Inputs:
            cluster_means (torch.Tensor): output from find_cluster_means
        Returns:
            reg_loss (float): computed regularization loss (see paper).
        """
        reg_loss = 0.0
        n_clusters, _ = cluster_means.shape
        for i in range(n_clusters):
            reg_loss += torch.norm(cluster_means[i, :] + 1e-8, p=self.norm)
        reg_loss /= float(n_clusters)
        return reg_loss

    def inter_cluster_loss(self, cluster_means, margin=0.2):
        """
        Implementation of distance loss in Discriminative Loss.
        Inputs:
            cluster_means (torch.Tensor): output from find_cluster_means
            margin (float/int): the magnitude of the margin delta_d in the paper.
            Think of it as the distance between each separate clusters in
            embedding space.
        Returns:
            inter_loss (float): computed cross-centroid distance loss (see paper).
            Factor of 2 is included for proper normalization.
        """
        inter_loss = 0.0
        n_clusters = len(cluster_means)
        if n_clusters < 2:
            # Inter-cluster loss is zero if there only one instance exists for
            # a semantic label.
            return 0.0
        else:
            for i, c1 in enumerate(cluster_means):
                for j, c2 in enumerate(cluster_means):
                    if i != j:
                        dist = torch.norm(c1 - c2 + 1e-8, p=self.norm)
                        hinge = torch.clamp(2.0 * margin - dist, min=0)
                        inter_loss += torch.pow(hinge, 2)
            inter_loss /= float((n_clusters - 1) * n_clusters)
            return inter_loss

    def get_per_class_probabilities(self, embeddings, margins, labels):
        """
        Computes binary foreground/background loss.
        """
        device = embeddings.device
        loss = 0.0
        smoothing_loss = 0.0
        centroids = self.find_cluster_means(embeddings, labels)
        inter_loss = self.inter_cluster_loss(centroids)
        n_clusters = len(centroids)
        cluster_labels = labels.unique(sorted=True)
        probs = torch.zeros(embeddings.shape[0]).float().to(device)
        accuracy = 0.0

        for i, c in enumerate(cluster_labels):
            index = labels == c
            mask = torch.zeros(embeddings.shape[0]).to(device)
            mask[index] = 1.0
            mask[~index] = 0.0
            sigma = torch.mean(margins[index], dim=0)
            dists = torch.sum(torch.pow(embeddings - centroids[i], 2), dim=1)
            p = torch.clamp(torch.exp(-dists / (2 * torch.pow(sigma, 2))), min=0, max=1)
            logits = torch.logit(p, eps=1e-6)
            # print(logits.shape)
            probs[index] = p[index]
            loss += lovasz_hinge_flat(logits, mask).mean()
            accuracy += float(iou_binary(p > 0.5, mask, per_image=False))
            sigma_detach = sigma.detach()
            smoothing_loss += torch.mean(
                torch.norm(margins[index].view(-1, 1) - sigma_detach, dim=1)
            )

        loss /= n_clusters
        smoothing_loss /= n_clusters
        accuracy /= n_clusters
        loss += inter_loss

        return loss, smoothing_loss, float(inter_loss), probs, accuracy

    def combine_multiclass(self, embeddings, margins, seediness, slabels, clabels):
        """
        Wrapper function for combining different components of the loss,
        in particular when clustering must be done PER SEMANTIC CLASS.

        NOTE: When there are multiple semantic classes, we compute the DLoss
        by first masking out by each semantic segmentation (ground-truth/prediction)
        and then compute the clustering loss over each masked point cloud.

        INPUTS:
            features (torch.Tensor): pixel embeddings
            slabels (torch.Tensor): semantic labels
            clabels (torch.Tensor): group/instance/cluster labels

        OUTPUT:
            loss_segs (list): list of computed loss values for each semantic class.
            loss[i] = computed DLoss for semantic class <i>.
            acc_segs (list): list of computed clustering accuracy for each semantic class.
        """
        loss = defaultdict(list)
        accuracy = defaultdict(float)
        semantic_classes = slabels.unique()
        for sc in semantic_classes:
            if int(sc) == 4:
                continue
            index = slabels == sc
            clabels_unique, _ = unique_label_torch(clabels[index])
            mask_loss, smoothing_loss, inter_loss, probs, acc = (
                self.get_per_class_probabilities(
                    embeddings[index], margins[index], clabels_unique
                )
            )
            prob_truth = probs.detach()
            seed_loss = self.l2loss(prob_truth, seediness[index].squeeze(1))
            total_loss = (
                self.embedding_weight * mask_loss
                + self.seediness_weight * seed_loss
                + self.smoothing_weight * smoothing_loss
            )
            loss["loss"].append(total_loss)
            loss["mask_loss"].append(float(self.embedding_weight * mask_loss))
            loss["seed_loss"].append(float(self.seediness_weight * seed_loss))
            loss["smoothing_loss"].append(float(self.smoothing_weight * smoothing_loss))
            loss["inter_loss"].append(float(self.inter_weight * inter_loss))
            loss["mask_loss_{}".format(int(sc))].append(float(mask_loss))
            loss["seed_loss_{}".format(int(sc))].append(float(seed_loss))
            accuracy["accuracy_{}".format(int(sc))] = acc

        return loss, accuracy


class MaskLovaszInterLoss2(MaskLovaszInterLoss):

    def __init__(self, cfg, name="spice_loss"):
        super(MaskLovaszInterLoss2, self).__init__(cfg, name)
        self.inter_weight = self.loss_config.get("inter_weight", 1.0)
        self.norm = 2
        self.seed_loss = torch.nn.BCEWithLogitsLoss()
        self.seed_threshold = self.loss_config.get("seed_threshold", 0.98)

    def get_per_class_probabilities(self, embeddings, margins, labels):
        """
        Computes binary foreground/background loss.
        """
        device = embeddings.device
        loss = 0.0
        smoothing_loss = 0.0
        sigma_reg_loss = 0.0
        centroids = self.find_cluster_means(embeddings, labels)
        # inter_loss = self.inter_cluster_loss(centroids)
        n_clusters = len(centroids)
        cluster_labels = labels.unique(sorted=True)
        probs = torch.zeros(embeddings.shape[0]).float().to(device)
        accuracy = 0.0

        for i, c in enumerate(cluster_labels):
            index = labels == c
            mask = torch.zeros(embeddings.shape[0]).to(device)
            mask[index] = 1
            mask[~index] = 0
            sigma = torch.mean(margins[index])
            dists = torch.sum(torch.pow(embeddings - centroids[i], 2), dim=1)
            p = torch.exp(-dists / (2 * torch.pow(sigma, 2) + 1e-8))
            probs[index] = p[index]
            loss += lovasz_hinge_flat(2 * p - 1, mask)
            accuracy += float(iou_binary(p > 0.5, mask, per_image=False))
            sigma_detach = sigma.detach()
            smoothing_loss += torch.mean(
                torch.norm(margins[index] - sigma_detach, dim=1)
            )
            sigma_reg_loss += torch.log(1.0 + sigma)

        loss /= n_clusters
        smoothing_loss /= n_clusters
        accuracy /= n_clusters
        sigma_reg_loss /= n_clusters
        loss += sigma_reg_loss

        return loss, smoothing_loss, probs, accuracy

    def combine_multiclass(self, embeddings, margins, seediness, slabels, clabels):
        """
        Wrapper function for combining different components of the loss,
        in particular when clustering must be done PER SEMANTIC CLASS.

        NOTE: When there are multiple semantic classes, we compute the DLoss
        by first masking out by each semantic segmentation (ground-truth/prediction)
        and then compute the clustering loss over each masked point cloud.

        INPUTS:
            features (torch.Tensor): pixel embeddings
            slabels (torch.Tensor): semantic labels
            clabels (torch.Tensor): group/instance/cluster labels

        OUTPUT:
            loss_segs (list): list of computed loss values for each semantic class.
            loss[i] = computed DLoss for semantic class <i>.
            acc_segs (list): list of computed clustering accuracy for each semantic class.
        """
        loss = defaultdict(list)
        accuracy = defaultdict(float)
        semantic_classes = slabels.unique()
        for sc in semantic_classes:
            if int(sc) == 4:
                continue
            index = slabels == sc
            mask_loss, smoothing_loss, probs, acc = self.get_per_class_probabilities(
                embeddings[index], margins[index], clabels[index]
            )
            prob_truth = probs.detach()
            target = (probs > self.seed_threshold).float()
            seed_loss = self.seed_loss(seediness[index].squeeze(1), target)
            total_loss = (
                self.embedding_weight * mask_loss
                + self.seediness_weight * seed_loss
                + self.smoothing_weight * smoothing_loss
            )
            loss["loss"].append(total_loss)
            loss["mask_loss"].append(float(self.embedding_weight * mask_loss))
            loss["seed_loss"].append(float(self.seediness_weight * seed_loss))
            loss["smoothing_loss"].append(float(self.smoothing_weight * smoothing_loss))
            # loss['inter_loss'].append(
            #     float(self.inter_weight * inter_loss))
            loss["mask_loss_{}".format(int(sc))].append(float(mask_loss))
            loss["seed_loss_{}".format(int(sc))].append(float(seed_loss))
            accuracy["accuracy_{}".format(int(sc))] = acc

        return loss, accuracy


class MaskLovaszInterBC(MaskLovaszInterLoss):

    def __init__(self, cfg, name="spice_loss"):
        super(MaskLovaszInterBC, self).__init__(cfg, name)
        self.inter_weight = self.loss_config.get("inter_weight", 1.0)
        self.norm = 2
        self.seed_loss = torch.nn.BCEWithLogitsLoss()
        self.seed_threshold = self.loss_config.get("seed_threshold", 0.98)

    def inter_cluster_loss(self, embeddings, margins, labels):
        vecs = torch.cat([embeddings, margins], dim=1)
        v = self.find_cluster_means(vecs, labels)
        mat = bhattacharyya_coeff_matrix(v, v)
        ind = torch.triu_indices(v.shape[0], v.shape[0], offset=1)
        inter = mat[ind[0], ind[1]]
        if inter.shape[0] == 0:
            return 0
        else:
            out = torch.mean(inter)
            return out

    def get_per_class_probabilities(self, embeddings, margins, labels):
        """
        Computes binary foreground/background loss.
        """
        device = embeddings.device
        loss = 0.0
        smoothing_loss = 0.0
        sigma_reg_loss = 0.0
        centroids = self.find_cluster_means(embeddings, labels)
        inter_loss = self.inter_cluster_loss(embeddings, margins, labels)
        n_clusters = len(centroids)
        cluster_labels = labels.unique(sorted=True)
        probs = torch.zeros(embeddings.shape[0]).float().to(device)
        accuracy = 0.0

        for i, c in enumerate(cluster_labels):
            index = labels == c
            mask = torch.zeros(embeddings.shape[0]).to(device)
            mask[index] = 1
            mask[~index] = 0
            sigma = torch.mean(margins[index])
            dists = torch.sum(torch.pow(embeddings - centroids[i], 2), dim=1)
            p = torch.exp(-dists / (2 * torch.pow(sigma, 2) + 1e-8))
            probs[index] = p[index]
            loss += lovasz_hinge_flat(2 * p - 1, mask)
            accuracy += float(iou_binary(p > 0.5, mask, per_image=False))
            sigma_detach = sigma.detach()
            smoothing_loss += torch.mean(
                torch.norm(margins[index] - sigma_detach, dim=1)
            )
            sigma_reg_loss += torch.log(1.0 + sigma)

        loss /= n_clusters
        smoothing_loss /= n_clusters
        accuracy /= n_clusters
        sigma_reg_loss /= n_clusters
        loss += sigma_reg_loss
        loss += inter_loss

        return loss, smoothing_loss, float(inter_loss), probs, accuracy

    def combine_multiclass(self, embeddings, margins, seediness, slabels, clabels):
        """
        Wrapper function for combining different components of the loss,
        in particular when clustering must be done PER SEMANTIC CLASS.

        NOTE: When there are multiple semantic classes, we compute the DLoss
        by first masking out by each semantic segmentation (ground-truth/prediction)
        and then compute the clustering loss over each masked point cloud.

        INPUTS:
            features (torch.Tensor): pixel embeddings
            slabels (torch.Tensor): semantic labels
            clabels (torch.Tensor): group/instance/cluster labels

        OUTPUT:
            loss_segs (list): list of computed loss values for each semantic class.
            loss[i] = computed DLoss for semantic class <i>.
            acc_segs (list): list of computed clustering accuracy for each semantic class.
        """
        loss = defaultdict(list)
        accuracy = defaultdict(float)
        semantic_classes = slabels.unique()
        for sc in semantic_classes:
            if int(sc) == 4:
                continue
            index = slabels == sc
            mask_loss, smoothing_loss, inter_loss, probs, acc = (
                self.get_per_class_probabilities(
                    embeddings[index], margins[index], clabels[index]
                )
            )
            prob_truth = probs.detach()
            seed_loss = self.l2loss(prob_truth, seediness[index].squeeze(1))
            total_loss = (
                self.embedding_weight * mask_loss
                + self.seediness_weight * seed_loss
                + self.smoothing_weight * smoothing_loss
            )
            loss["loss"].append(total_loss)
            loss["mask_loss"].append(float(self.embedding_weight * mask_loss))
            loss["seed_loss"].append(float(self.seediness_weight * seed_loss))
            loss["smoothing_loss"].append(float(self.smoothing_weight * smoothing_loss))
            loss["inter_loss"].append(float(self.inter_weight * inter_loss))
            loss["mask_loss_{}".format(int(sc))].append(float(mask_loss))
            loss["seed_loss_{}".format(int(sc))].append(float(seed_loss))
            accuracy["accuracy_{}".format(int(sc))] = acc

        return loss, accuracy


class MaskFocalLoss(MaskBCELoss2):
    """
    Spatial Embeddings Loss with trainable center of attention.
    """

    def __init__(self, cfg, name="spice_loss"):
        super(MaskFocalLoss, self).__init__(cfg, name)
        self.bceloss = FocalLoss(logits=False)

    def get_per_class_probabilities(self, embeddings, margins, labels, coords):
        """
        Computes binary foreground/background loss.
        """
        device = embeddings.device
        loss = 0.0
        smoothing_loss = 0.0
        centroids = self.find_cluster_means(embeddings, labels)
        n_clusters = len(centroids)
        cluster_labels = labels.unique(sorted=True)
        probs = torch.zeros(embeddings.shape[0]).float().to(device)
        acc = 0.0

        for i, c in enumerate(cluster_labels):
            index = labels == c
            mask = torch.zeros(embeddings.shape[0]).to(device)
            mask[index] = 1.0
            mask[~index] = 0.0
            sigma = torch.mean(margins[index], dim=0)
            dists = torch.sum(torch.pow(embeddings - centroids[i], 2), dim=1)
            logits = dists / (2 * torch.pow(sigma, 2))
            p = torch.clamp(torch.exp(-dists / (2 * torch.pow(sigma, 2))), min=0, max=1)
            probs[index] = p[index]
            loss += self.bceloss(p, mask)
            acc += iou_binary(p > 0.5, mask, per_image=False)
            sigma_detach = sigma.detach()
            smoothing_loss += torch.sum(torch.pow(margins[index] - sigma_detach, 2))

        loss /= n_clusters
        smoothing_loss /= n_clusters
        acc /= n_clusters

        return loss, smoothing_loss, probs, acc


class MultiVariateLovasz(MaskLovaszInterLoss):

    def __init__(self, cfg, name="spice_loss"):
        super(MultiVariateLovasz, self).__init__(cfg, name)

    def get_per_class_probabilities(self, embeddings, margins, labels, coords):
        """
        Computes binary foreground/background loss.
        """
        device = embeddings.device
        loss = 0.0
        smoothing_loss = 0.0
        centroids = self.find_cluster_means(embeddings, labels)
        inter_loss = self.inter_cluster_loss(centroids)
        reg_loss = self.regularization(centroids)
        n_clusters = len(centroids)
        cluster_labels = labels.unique(sorted=True)
        probs = torch.zeros(embeddings.shape[0]).float().to(device)
        accuracy = 0.0

        if embeddings.shape[0] < 2:
            return 0, 0, 0, 0, 0

        for i, c in enumerate(cluster_labels):
            index = labels == c
            mask = torch.zeros(embeddings.shape[0]).to(device)
            mask[index] = 1
            mask[~index] = 0
            sigma = torch.mean(margins[index], dim=0)
            f = multivariate_kernel(centroids[i], sigma)
            p = f(embeddings)
            probs[index] = p[index]
            loss += lovasz_hinge_flat(2 * p - 1, mask)
            accuracy += iou_binary(p > 0.5, mask, per_image=False)
            sigma_detach = sigma.detach()
            smoothing_loss += torch.sum(torch.pow(margins[index] - sigma_detach, 2))

        loss /= n_clusters
        smoothing_loss /= n_clusters
        accuracy /= n_clusters
        loss += inter_loss
        loss += reg_loss / n_clusters

        return loss, smoothing_loss, inter_loss, probs, accuracy



"""Draw reconstruction output-level objects"""

import numpy as np
from plotly import graph_objs as go

from spine.geo import GeoManager
from spine.utils.globals import PID_LABELS, SHAPE_LABELS, TRACK_SHP

from .arrow import scatter_arrows
from .cluster import scatter_clusters
from .geo import GeoDrawer
from .layout import HIGH_CONTRAST_COLORS, PLOTLY_COLORS_WGRAY, dual_figure3d, layout3d
from .lite import scatter_lite
from .point import scatter_points

__all__ = ["Drawer"]


class Drawer:
    """Handles drawing the true/reconstructed output.

    This class is given the entire input/output dictionary from one entry and
    provides functions to represent the output.
    """

    # List of recognized object types
    _obj_types = ("fragments", "particles", "interactions")

    # List of recognized drawing modes
    _draw_modes = ("reco", "truth", "both", "all")

    # List of known point modes for true objects and their corresponding keys
    _point_modes = (
        ("points", "points_label"),
        ("points_adapt", "points"),
        ("points_g4", "points_g4"),
    )

    # List of known deposition modes for true particles and their corresponding keys
    _dep_modes = (
        ("depositions", "depositions_label"),
        ("depositions_q", "depositions_q_label"),
        ("depositions_adapt", "depositions_label_adapt"),
        ("depositions_adapt_q", "depositions"),
        ("depositions_g4", "depositions_g4"),
    )

    # List of known source modes for true particles and their corresponding keys
    _source_modes = (("sources", "sources_label"), ("sources_adapt", "sources"))

    def __init__(
        self,
        data,
        draw_mode="both",
        truth_point_mode="points",
        split_scene=True,
        geo=None,
        detector_coords=True,
        lite=False,
        **kwargs,
    ):
        """Initialize the drawer attributes

        Parameters
        ----------
        data : dict
            Dictionary of data products
        draw_mode : str, default 'both'
            Drawing mode, one of 'reco', 'truth' or 'both'
        truth_point_mode : str, optional
            If specified, tells which attribute of the :class:`TruthFragment`,
            :class:`TruthParticle` or :class:`TruthInteraction` object to use
            to fetch its point coordinates
        split_scene : bool, default True
            If True and when drawing both reconstructed and truth information,
            split the traces between two separate scenes
        geo : Geometry, optional
            Geometry object used to draw detector elements
        detector_coords : bool, default True
            Whether the object coordinates are expressed in detector coordinates
        lite : bool, default False
            If `True`, load the objects in their lite version (no space points)
        **kwargs : dict, optional
            Additional arguments to pass to the :func:`layout3d` function
        """
        # Store the data used to draw the reconstruction output
        self.data = data

        # Set up the list of prefixes to access
        assert draw_mode in self._draw_modes, (
            f"`mode` not recognized: {draw_mode}. Must be one of "
            f"{self._draw_modes}."
        )

        self.prefixes = []
        if draw_mode != "truth":
            self.prefixes.append("reco")
        if draw_mode != "reco":
            self.prefixes.append("truth")

        # Build a list of supported objects
        self.supported_objs = []
        for mode in ["reco", "truth"]:
            for obj_type in self._obj_types:
                self.supported_objs.append(f"{mode}_{obj_type}")

        # Set up the truth point mode
        assert truth_point_mode in self.point_modes, (
            "The `truth_point_mode` argument must be one of "
            f"{self.point_modes.keys()}. Got `{truth_point_mode}` instead."
        )
        self.truth_point_mode = truth_point_mode
        self.truth_point_key = self.point_modes[self.truth_point_mode]
        self.truth_index_mode = truth_point_mode.replace("points", "index")

        # If detector information is provided, initialize the geometry drawer
        self.geo, self.geo_drawer = None, None
        self.meta = data.get("meta", None)
        if geo is not None or GeoManager.is_initialized():
            self.geo = geo if geo is not None else GeoManager.get_instance()
            self.geo_drawer = GeoDrawer(geo=self.geo, detector_coords=detector_coords)
            self.geo = self.geo_drawer.geo

        # Store whether to draw in lite mode
        self.lite = lite

        # Store the layout information
        self.split_scene = split_scene
        self.detector_coords = detector_coords
        self.meta = self.meta if self.geo is None else None
        self.layout_kwargs = kwargs

    @property
    def point_modes(self):
        """Dictionary which makes the correspondance between the name of a true
        object point attribute with the underlying point tensor it points to.

        Returns
        -------
        Dict[str, str]
            Dictionary of (attribute, key) mapping for point coordinates
        """
        return dict(self._point_modes)

    @property
    def dep_modes(self):
        """Dictionary which makes the correspondance between the name of a true
        object deposition attribute with the underlying deposition array it points to.

        Returns
        -------
        Dict[str, str]
            Dictionary of (attribute, key) mapping for point depositions
        """
        return dict(self._dep_modes)

    @property
    def source_modes(self):
        """Dictionary which makes the correspondance between the name of a true
        object source attribute with the underlying source array it points to.

        Returns
        -------
        Dict[str, str]
            Dictionary of (attribute, key) mapping for point source
        """
        return dict(self._source_modes)

    def get_index(self, obj):
        """Get a certain pre-defined index attribute of an object.

        The :class:`TruthFragment`, :class:`TruthParticle` and
        :class:`TruthInteraction` objects index are obtained using the
        `truth_index_mode` attribute of the class.

        Parameters
        ----------
        obj : Union[FragmentBase, ParticleBase, InteractionBase]
            Fragment, Particle or Interaction object

        Results
        -------
        np.ndarray
           (N) Object index
        """
        if not obj.is_truth:
            return obj.index
        else:
            return getattr(obj, self.truth_index_mode)

    def get(
        self,
        obj_type,
        attr=None,
        color_attr=None,
        draw_raw=False,
        draw_end_points=False,
        draw_directions=False,
        draw_vertices=False,
        draw_flashes=False,
        matched_flash_only=True,
        draw_crthits=False,
        matched_crthit_only=True,
        synchronize=False,
        titles=None,
        split_traces=False,
    ):
        """Draw the requested object type with the requested mode.

        Parameters
        ----------
        obj_type : str
            Name of the object type to draw (one of 'fragments', 'particles' or
            'interactions'
        attr : Union[str, List[str]], optional
            Name of list of names of attributes to draw
        color_attr : str, optional
            Name of the attribute to use to determine the color
        draw_raw : bool, default False
            If `True`, add a trace which corresponds to the raw depositions
        draw_end_points : bool, default False
            If `True`, draw the fragment or particle end points
        draw_directions : bool, default False
            If `True`, draw the fragment or particle start directions
        draw_vertices : bool, default False
            If `True`, draw the interaction vertices
        draw_flashes : bool, default False
            If `True`, draw the flashes
        matched_flash_only : bool, default True
            If `True`, only flashes matched to interactions are drawn
        draw_crthits : bool, default False
            If `True`, draw the CRT hits
        matched_crthit_only : bool, default True
            If `True`, only CRT hits matched to interactions are drawn
        synchronize : bool, default False
            If `True`, matches the camera position/angle of one plot to the other
        titles : List[str], optional
            Titles of the two scenes (only relevant for split_scene=True)
        split_traces : bool, default False
            If `True`, one trace is produced for each object

        Returns
        -------
        go.Figure
            Figure containing all the necessary information to draw
        """
        # Check that what is to be drawn is a known object type that is provided
        assert obj_type in self._obj_types, (
            f"Object type not recognized: {obj_type}. Must be one of "
            f"{self._obj_types}."
        )

        # Fetch the objects
        traces = {}
        for prefix in self.prefixes:
            obj_name = f"{prefix}_{obj_type}"
            assert obj_name in self.data, (
                f"Must provide `{obj_name}` in the data products to draw " "them."
            )
            traces[prefix] = self._object_traces(
                obj_name, attr, color_attr, split_traces
            )

        # Fetch the raw depositions, if requested
        if draw_raw:
            for prefix in self.prefixes:
                traces[prefix] = self._raw_trace() + traces[prefix]

        # Fetch the end points, if requested
        if draw_end_points:
            assert (
                obj_type != "interactions"
            ), "Interactions do not have end point attributes."
            for prefix in self.prefixes:
                obj_name = f"{prefix}_{obj_type}"
                traces[prefix] += self._start_point_trace(obj_name, split_traces)
                traces[prefix] += self._end_point_trace(obj_name, split_traces)

        # Fetch the directions, if requested
        if draw_directions:
            assert (
                obj_type != "interactions"
            ), "Interactions do not have direction attributes."
            for prefix in self.prefixes:
                obj_name = f"{prefix}_{obj_type}"
                traces[prefix] += self._direction_trace(obj_name, split_traces)

        # Fetch the vertices, if requested
        if draw_vertices:
            for prefix in self.prefixes:
                obj_name = f"{prefix}_interactions"
                assert (
                    obj_name in self.data
                ), "Must provide interactions to draw their vertices."
                traces[prefix] += self._vertex_trace(obj_name, split_traces)

        # Fetch the flashes, if requested
        show_optical = False
        if draw_flashes:
            assert (
                "flashes" in self.data
            ), "Must provide the `flashes` objects to draw them."
            show_optical = True
            for prefix in self.prefixes:
                obj_name = f"{prefix}_interactions"
                assert (
                    obj_name in self.data
                ), "Must provide interactions to draw matched flashes."
                traces[prefix] += self._flash_trace(obj_name, matched_flash_only)

        # Fetch the CRT hits, if requested
        show_crt = False
        if draw_crthits:
            assert (
                "crthits" in self.data
            ), "Must provide the `crthits` objects to draw them."
            show_crt = True
            for prefix in self.prefixes:
                obj_name = f"{prefix}_{obj_type}"
                traces[prefix] += self._crt_trace(obj_name, matched_crthit_only)

        # Add the TPC traces, if available
        if self.geo_drawer is not None:
            if self.prefixes and self.split_scene:
                for prefix in self.prefixes:
                    traces[prefix] += self.geo_drawer.tpc_traces(meta=self.meta)
            else:
                traces[self.prefixes[-1]] += self.geo_drawer.tpc_traces(meta=self.meta)

        # Initialize the layout based on what needs to be shown
        layout = layout3d(
            geo=self.geo,
            use_geo=self.geo is not None,
            meta=self.meta,
            detector_coords=self.detector_coords,
            show_optical=show_optical,
            show_crt=show_crt,
            **self.layout_kwargs,
        )

        # Initialize the figure, return
        if len(self.prefixes) > 1 and self.split_scene:
            if titles is None:
                titles = [f"Reconstructed {obj_type}", f"Truth {obj_type}"]
            figure = dual_figure3d(
                traces["reco"],
                traces["truth"],
                layout=layout,
                synchronize=synchronize,
                titles=titles,
            )

        else:
            assert titles is None, (
                "Providing titles does not do anything when split_scene " "is False."
            )
            all_traces = []
            for trace_group in traces.values():
                all_traces += trace_group
            figure = go.Figure(all_traces, layout=layout)

        return figure

    def _object_traces(self, obj_name, attr, color_attr, split_traces):
        """Draw a specific object.

        Parameters
        ----------
        obj_name : str
            Name of the objects to be represented
        attr : Union[str, List[str]]
            Attribute name(s) used to set the color/hovertext
        color_attr : str
            Name of the attribute to use to determine the color
        split_traces : bool
            If `True`, one trace is produced for each object

        Returns
        -------
        List[plotly.graph_objs.Scatter3d]
            List of traces, one per object being drawn up
        """
        # Get the colors
        if attr is not None and isinstance(attr, str):
            attr = [attr]
        color_dict = self._object_colors(obj_name, attr, color_attr, split_traces)

        # Fetch object representations
        if not self.lite:
            # Check that the object has points to be drawn
            point_key = self.truth_point_key if "truth" in obj_name else "points"
            assert point_key in self.data, (
                f"The `{point_key}` attribute must be provided if the full "
                f"version of the `{obj_name}` objects is to be drawn."
            )

            # Fetch the points and the clusters
            points = self.data[point_key]
            clusts = [self.get_index(obj) for obj in self.data[obj_name]]

            # Initialize traces
            traces = scatter_clusters(
                points,
                clusts,
                single_trace=not split_traces,
                shared_legend=not split_traces,
                **color_dict,
            )

        else:
            # Initialize traces
            traces = scatter_lite(
                self.data[obj_name],
                shared_legend=not split_traces,
                **color_dict,
            )

        return traces

    def _object_colors(self, obj_name, attrs, color_attr, split_traces):
        """Provides an appropriate colorscale and range for a given attribute.

        Parameters
        ----------
        obj_name : str
            Name of the object to draw
        attrs : Optional[List[str]]
            Attribute name(s) used to set the color/hovertext
        color_attr : Optional[str]
            Name of the attribute to use to determine the color
        split_traces : bool
            If `True`, one trace is produced for each object

        Returns
        -------
        dict
            Dictionary of color parameters (colorscale, cmin, cmax)
        """
        # Define the name of the trace group
        name = " ".join(obj_name.split("_")).capitalize()
        if split_traces:
            name = name[:-1]

        # Make sure the color attribute is included in the hovertext attributes
        if color_attr is not None:
            assert color_attr in attrs, (
                "The attribute used to define the color scale must be "
                "included in the list of hovertext attributes."
            )

        # Make sure that the geometry is provided if needed
        if any(self._is_sources(attr) for attr in attrs or []):
            assert self.geo is not None, (
                "Provide detector name/geometry if the TPC "
                "sources are to be displayed."
            )

        # Initialize hovertext per object
        obj_type = obj_name.split("_")[-1][:-1].capitalize()
        count = len(self.data[obj_name])
        hovertext = [[f"{obj_type} {i}"] for i in range(count)]
        has_long_form = False
        if attrs is not None and any(self._is_long_form(attr) for attr in attrs):
            if self.lite:
                raise ValueError(
                    "Long-form attributes are not available when drawing "
                    "lite-version objects."
                )
            point_key = self.truth_point_key if "truth" in obj_name else "points"
            hovertext = [ht * len(self.data[point_key]) for ht in hovertext]
            has_long_form = True

        # Fetch the list of colors and hovertexts
        color = np.arange(len(self.data[obj_name]))
        if color_attr is None:
            color_attr = "id"

        if attrs is not None:
            # Fetch hover information for each of the requested attributes
            for attr in attrs:
                # Get the atribute name in display format
                attr_name = " ".join(attr.split("_")).capitalize()

                # If it is a true deposition/source attribute, check that it matches
                # the point mode that is being used to draw the true objects
                if "truth" in obj_name:
                    if self._is_depositions(attr):
                        prefix = self.truth_point_mode.replace("points", "depositions")
                        assert attr.startswith(prefix), (
                            f"Points mode {self.truth_point_mode} and deposition "
                            f"mode {attr} are incompatible."
                        )
                    if self._is_sources(attr):
                        ref_name = self.truth_point_mode.replace("points", "sources")
                        assert attr == ref_name, (
                            f"Points mode {self.truth_point_mode} and source "
                            f"mode {attr} are incompatible."
                        )

                # Get the list of values
                values = [getattr(obj, attr) for obj in self.data[obj_name]]
                if self._is_sources(attr):
                    assert self.geo is not None  # For the linter's sake
                    values = [self.geo.get_sources(v) for v in values]

                # Get the colors, if needed
                if len(attrs) == 1 or attr == color_attr:
                    if not self._is_sources(attr):
                        color = values
                    else:
                        assert self.geo is not None, (
                            "Provide detector name/geometry if the TPC "
                            "sources are to be displayed."
                        )
                        color = [self.geo.get_chambers(v) for v in values]

                # Get the hovertext for this attribute
                if self._is_depositions(attr):
                    for i, ht in enumerate(hovertext):
                        hovertext[i] = [
                            ht[j] + self._dep_tostr(v) for j, v in enumerate(values[i])
                        ]
                elif self._is_sources(attr):
                    for i, ht in enumerate(hovertext):
                        hovertext[i] = [
                            ht[j] + self._src_tostr(v) for j, v in enumerate(values[i])
                        ]
                else:
                    for i, ht in enumerate(hovertext):
                        val_str = self._tostr(attr_name, values[i])
                        hovertext[i] = [
                            ht[j] + val_str for j in range(len(hovertext[i]))
                        ]

            # If there is only one attribute and no color attribute specified, use it
            if color_attr is None and len(attrs) == 1:
                color_attr = attrs[0]

        # Flatten hovertext if no long-form attributes are used
        if not has_long_form:
            hovertext = [ht[0] for ht in hovertext]

        # Set up the appropriate color scheme
        if self._is_depositions(color_attr):
            # Continuous values shared between objects
            dep_mode = (
                self.dep_modes[color_attr] if "truth" in obj_name else "depositions"
            )
            colorscale = "Inferno"
            cmin = 0.0
            cmax = 2 * np.median(self.data[dep_mode])

        elif self._is_sources(color_attr):
            # Variable-length descrete values
            assert self.geo is not None  # For the linter's sake
            count = self.geo.tpc.num_chambers
            colorscale = HIGH_CONTRAST_COLORS
            if count == 0:
                colorscale = None
            elif count == 1:
                colorscale = [colorscale[0]] * 2  # Avoid length 0 colorscale
            elif count <= len(colorscale):
                colorscale = colorscale[:count]
            else:
                repeat = (count - 1) // len(colorscale) + 1
                colorscale = np.tile(colorscale, repeat)[:count]

            cmin = 0
            cmax = count - 1

        elif color_attr.startswith("is_"):
            # Boolean
            color = np.array(color, dtype=np.int32)
            colorscale = PLOTLY_COLORS_WGRAY[1:3]
            cmin = 0
            cmax = 1

        elif color_attr == "shape" or color_attr == "pid":
            # Fixed length values, with potentially invalid values
            ref = SHAPE_LABELS if color_attr == "shape" else PID_LABELS
            num_classes = len(ref)
            colorscale = PLOTLY_COLORS_WGRAY[: num_classes + 1]
            cmin = -1
            cmax = num_classes - 1

        elif color_attr.endswith("id"):
            # Variable-lengh discrete values
            unique, color = np.unique(color, return_inverse=True)
            colorscale = HIGH_CONTRAST_COLORS
            count = len(unique)
            if count == 0:
                colorscale = None
            elif count == 1:
                colorscale = [colorscale[0]] * 2  # Avoid length 0 colorscale
            elif count <= len(colorscale):
                colorscale = colorscale[:count]
            else:
                repeat = (count - 1) // len(colorscale) + 1
                colorscale = np.tile(colorscale, repeat)[:count]

            cmin = 0
            cmax = count - 1

        else:
            raise ValueError(f"Color attribute not supported: {color_attr}.")

        return {
            "color": color,
            "hovertext": hovertext,
            "name": name,
            "colorscale": colorscale,
            "cmin": cmin,
            "cmax": cmax,
        }

    def _raw_trace(self):
        """Draws the raw input image (pre-reconstruction).

        Returns
        -------
        List[plotly.graph_objs.Scatter3d]
            List of one trace containing the input to the reconstruction
        """
        # Throw if drawing in lite mode
        if self.lite:
            raise RuntimeError("Cannot draw raw input in lite mode.")

        # Fetch the input attributes
        points = self.data["points"]
        deps = self.data["depositions"]

        # Fetch the colorscale limits
        cmin = 0.0
        cmax = 2 * np.median(deps) if len(deps) else 1.0

        return scatter_points(
            points,
            color=deps,
            cmin=cmin,
            cmax=cmax,
            colorscale="Inferno",
            name="Raw input",
        )

    def _start_point_trace(
        self,
        obj_name,
        split_traces,
        color="black",
        markersize=7,
        marker_symbol="circle",
        **kwargs,
    ):
        """Scatters the start points of the requested object type.

        Parameters
        ----------
        obj_name : str
            Name of the object to draw
        split_traces : bool
            If `True`, one trace is produced for each object
        color : Union[str, np.ndarray], default 'black'
            Color of markers/lines or (N) list of color of markers/lines
        markersize : float, default 7
            Marker size
        marker_symbol : float, default 'circle'
            Marker style
        **kwargs : dict, optional
            Additional parameters to pass

        Returns
        -------
        list
            List of start point traces
        """
        return self._point_trace(
            obj_name,
            "start_point",
            split_traces,
            color=color,
            markersize=markersize,
            marker_symbol=marker_symbol,
            **kwargs,
        )

    def _end_point_trace(
        self,
        obj_name,
        split_traces,
        color="black",
        markersize=7,
        marker_symbol="circle-open",
        **kwargs,
    ):
        """Scatters the end points of the requested object type.

        Parameters
        ----------
        obj_name : str
            Name of the object to draw
        split_traces : bool
            If `True`, one trace is produced for each object
        color : Union[str, np.ndarray], default 'black'
            Color of markers/lines or (N) list of color of markers/lines
        markersize : float, default 7
            Marker size
        marker_symbol : float, default 'circle-open'
            Marker style
        **kwargs : dict, optional
            Additional parameters to pass

        Returns
        -------
        list
            List of end point traces
        """
        return self._point_trace(
            obj_name,
            "end_point",
            split_traces,
            color=color,
            markersize=markersize,
            marker_symbol=marker_symbol,
            **kwargs,
        )

    def _vertex_trace(
        self,
        obj_name,
        split_traces,
        vertex_attr="vertex",
        color="green",
        markersize=10,
        marker_symbol="diamond",
        **kwargs,
    ):
        """Scatters the vertex of the requested object type.

        Parameters
        ----------
        obj_name : str
            Name of the object to draw
        split_traces : bool
            If `True`, one trace is produced for each object
        color : Union[str, np.ndarray], default 'green'
            Color of markers/lines or (N) list of color of markers/lines
        markersize : float, default 10
            Marker size
        marker_symbol : float, default 'circle-open'
            Marker style
        **kwargs : dict, optional
            Additional parameters to pass

        Returns
        -------
        list
            List of vertex point traces
        """
        return self._point_trace(
            obj_name,
            vertex_attr,
            split_traces,
            color=color,
            markersize=markersize,
            marker_symbol=marker_symbol,
            **kwargs,
        )

    def _point_trace(self, obj_name, point_attr, split_traces, **kwargs):
        """Scatters a set of discrete points per object instance.

        Parameters
        ----------
        obj_name : str
            Name of the object to draw
        point_attr : str
            Name of the attribute specifying end point to draw
        split_traces : bool
            If `True`, one trace is produced for each object
        **kwargs : dict, optional
            List of additional arguments to pass to :func:`scatter_points`

        Returns
        -------
        list
            List of point traces
        """
        # Define the name of the trace
        name = (
            " ".join(obj_name.split("_")).capitalize()[:-1]
            + " "
            + " ".join(point_attr.split("_"))
        )

        # Fetch the particular end point of each object
        obj_type = obj_name.split("_")[-1][:-1].capitalize()
        point_list, hovertext, idxs = [], [], []
        for i, obj in enumerate(self.data[obj_name]):
            # If it is an end point, skip if the object is not a track
            if point_attr == "end_point" and obj.shape != TRACK_SHP:
                continue

            # Skip empty true objects
            if obj.is_truth and not len(getattr(obj, self.truth_index_mode)) > 0:
                continue

            # Append the particular end point of this object and the label
            point_list.append(getattr(obj, point_attr))
            hovertext.append(f"{obj_type} {i} " + " ".join(point_attr.split("_")))
            idxs.append(i)

        points = np.empty((0, 3))
        if point_list:
            points = np.vstack(point_list)

        if not split_traces:
            traces = scatter_points(
                points, hovertext=np.array(hovertext), name=name, **kwargs
            )

        else:
            traces = []
            for i, point in enumerate(point_list):
                traces += scatter_points(
                    point[None, :],
                    hovertext=hovertext[i],
                    name=f"{name} {idxs[i]}",
                    **kwargs,
                )

        return traces

    def _direction_trace(self, obj_name, split_traces, color="black", **kwargs):
        """Scatters a set of discrete points per object instance.

        Parameters
        ----------
        obj_name : str
            Name of the object to draw
        split_traces : bool
            If `True`, one trace is produced for each object
        color : Union[str, np.ndarray], default 'black'
            Color of markers/lines or (N) list of color of markers/lines
        **kwargs : dict, optional
            List of additional arguments to pass to :func:`scatter_arrows`

        Returns
        -------
        list
            List of point traces
        """
        # Define the name of the trace
        name = " ".join(obj_name.split("_")).capitalize()[:-1] + " directions"

        # Fetch the direction of each object
        obj_type = obj_name.split("_")[-1][:-1].capitalize()
        point_list, dir_list, hovertext, idxs = [], [], [], []
        for i, obj in enumerate(self.data[obj_name]):
            # Skip empty true objects
            if obj.is_truth and not len(getattr(obj, self.truth_index_mode)) > 0:
                continue

            # Append the direction of this object and the label
            point_list.append(obj.start_point)
            dir_list.append(obj.start_dir)
            hovertext.append(f"{obj_type} {i} direction")
            idxs.append(i)

        points, dirs = np.empty((0, 3)), np.empty((0, 3))
        if point_list:
            points = np.vstack(point_list)
            dirs = np.vstack(dir_list)

        if not split_traces:
            traces = scatter_arrows(
                points,
                dirs,
                hovertext=np.array(hovertext),
                name=name,
                color=color,
                **kwargs,
            )

        else:
            traces = []
            for i, (point, start_dir) in enumerate(zip(point_list, dir_list)):
                traces += scatter_arrows(
                    point[None, :],
                    start_dir[None, :],
                    color=color,
                    hovertext=hovertext[i],
                    name=f"{name} {idxs[i]}",
                    **kwargs,
                )

        return traces

    def _flash_trace(self, obj_name, matched_only, **kwargs):
        """Draw the cumulative PEs of flashes.

        Parameters
        ----------
        obj_name : str
            Name of the object to draw
        matched_only : bool
            If `True`, only flashes matched to interactions are drawn
        **kwargs : dict, optional
            List of additional arguments to pass to :func:`optical_traces`

        Returns
        -------
        list
            List of optical detector traces
        """
        # If there was no geometry provided by the user, nothing to do here
        assert (
            self.geo_drawer is not None
        ), "Cannot draw optical detectors without geometry information."

        # Check that there are optical detectors to draw
        assert (
            self.geo is not None and self.geo.optical is not None
        ), "This geometry does not have optical detectors to draw."

        # Define the name of the trace
        name = " ".join(obj_name.split("_")).capitalize()[:-1] + " flashes"

        # Find the list of flash IDs to draw
        if matched_only:
            flash_ids = []
            for inter in self.data[obj_name]:
                if inter.is_flash_matched:
                    flash_ids.extend(inter.flash_ids)
        else:
            flash_ids = np.arange(len(self.data["flashes"]))

        # Sum values from each flash to build a a global color scale
        color = np.zeros(self.geo.optical.num_detectors)
        op_det_ids = self.geo.optical.det_ids
        for flash_id in flash_ids:
            flash = self.data["flashes"][flash_id]
            pe_per_ch = flash.pe_per_ch
            if not self.geo.optical.global_index:
                op_det_ids = self.geo.optical.volumes[flash.volume_id].det_ids
            if op_det_ids is not None:
                pe_per_ch = np.bincount(op_det_ids, weights=pe_per_ch)
            index = self.geo.optical.volume_index(flash.volume_id)
            color[index] += pe_per_ch

        # Return the set of optical detectors with a color scale
        return self.geo_drawer.optical_traces(
            meta=self.meta,
            color=color,
            zero_supress=True,
            colorscale="Inferno",
            name=name,
            **kwargs,
        )

    def _crt_trace(self, obj_name, matched_only, **kwargs):
        """Draw the CRT planes and the hits.

        Parameters
        ----------
        obj_name : str
            Name of the object to draw
        matched_only : bool
            If `True`, only CRT hits matched to interactions are drawn
        **kwargs : dict, optional
            List of additional arguments to pass to :func:`optical_traces`

        Returns
        -------
        list
            List of optical detector traces
        """
        # If there was no geometry provided by the user, nothing to do here
        assert (
            self.geo is not None and self.geo_drawer is not None
        ), "Cannot draw CRT detectors without geometry information."

        # Check that there are CRT planes to draw
        assert (
            self.geo.crt is not None
        ), "This geometry does not have CRT planes to draw."

        # Define the names of the traces
        name_pl = " ".join(obj_name.split("_")).capitalize()[:-1] + " CRT planes"
        name_hits = " ".join(obj_name.split("_")).capitalize()[:-1] + " CRT hits"

        # Fetch CRT hits. Restrict to matched hits, if requested
        crthits = self.data["crthits"]
        if matched_only:
            crt_ids = []
            for inter in self.data[obj_name]:
                if inter.is_crt_matched:
                    crt_ids.extend(inter.crt_ids)
            crt_ids = np.unique(crt_ids)
            crthits = [crthits[idx] for idx in crt_ids]

        # Identify which of the CRT planes were hit (to know what to draw)
        det_ids = [self.geo.crt.get_plane_id(hit.center, hit.plane) for hit in crthits]
        unique_det_ids = np.unique(det_ids)

        # Initialize the hovertext for the planes and hits
        hovertext_pl, hovertext_hits = [], []
        for i, det_id in enumerate(unique_det_ids):
            hovertext_pl.append(f"CRT Plane {det_id}")
        for i, hit in enumerate(crthits):
            hovertext_hits.append(f"CRT hit {hit.id}<br>CRT Plane ID: {det_ids[i]}")

        # Initialize the CRT plane traces
        traces = self.geo_drawer.crt_traces(
            meta=self.meta,
            draw_ids=unique_det_ids,
            hovertext=hovertext_pl,
            name=name_pl,
        )

        # Build a scatter plot of CRT hits
        points = np.empty((0, 3))
        if crthits:
            points = np.vstack([hit.center for hit in crthits])

        traces += scatter_points(
            points,
            color="gray",
            markersize=5,
            hovertext=hovertext_hits,
            name=name_hits,
            **kwargs,
        )

        return traces

    def _is_long_form(self, attr):
        """Check if an attribute is a long-form attribute (depositions or sources).

        Parameters
        ----------
        attr : str
            Object attribute to check

        Returns
        -------
        bool
            `True` is the attribute is a long-form attribute
        """
        return self._is_depositions(attr) or self._is_sources(attr)

    @staticmethod
    def _is_depositions(attr):
        """Check if an attribute represents one deposition value per point.

        Parameters
        ----------
        attr : str
            Object attribute to check

        Returns
        -------
        bool
            `True` is the attribute is a deposition attribute
        """
        return attr.startswith("depositions") and not attr.endswith("sum")

    @staticmethod
    def _is_sources(attr):
        """Check if an attribute represents one source value per point.

        Parameters
        ----------
        attr : str
            Object attribute to check

        Returns
        -------
        bool
            `True` is the attribute is a source attribute
        """
        return attr.startswith("sources")

    def _tostr(self, attr_name, value):
        """Convert a scalar attribute value to a human-readable string.

        Parameters
        ----------
        attr_name : str
            Name of the attribute
        value : Any
            Attribute value

        Returns
        -------
        str
            Human-readable string
        """
        return f"<br>{attr_name}: {value}"

    def _dep_tostr(self, value):
        """Convert a deposition value to a human-readable string.

        Parameters
        ----------
        value : str
            Deposition value

        Returns
        -------
        str
            Human-readable string
        """
        return f"<br>Deposition: {value:0.3f}"

    def _src_tostr(self, value):
        """Convert a source value to a human-readable string.

        Parameters
        ----------
        value : str
            Source value

        Returns
        -------
        str
            Human-readable string
        """
        return f"<br>Module, TPC: {value[0]:d}, {value[1]:d}"
